package solutions.bitbadger.documents

import kotlin.jvm.Throws
import kotlin.reflect.full.*
import kotlin.reflect.jvm.isAccessible

/**
 * Strategies for automatic document IDs
 */
enum class AutoId {
    /** No automatic IDs will be generated */
    DISABLED,

    /** Generate a `MAX`-plus-1 numeric ID */
    NUMBER,

    /** Generate a `UUID` string ID */
    UUID,

    /** Generate a random hex character string ID */
    RANDOM_STRING;

    companion object {

        /**
         * Generate a `UUID` string
         *
         * @return A `UUID` string
         */
        @JvmStatic
        fun generateUUID(): String =
            java.util.UUID.randomUUID().toString().replace("-", "")

        /**
         * Generate a string of random hex characters
         *
         * @param length The length of the string (optional; defaults to configured length)
         * @return A string of random hex characters of the requested length
         */
        @JvmStatic
        fun generateRandomString(length: Int? = null): String =
            (length ?: Configuration.idStringLength).let { len ->
                kotlin.random.Random.nextBytes((len + 2) / 2)
                    .joinToString("") { String.format("%02x", it) }
                    .substring(0, len)
            }

        /**
         * Determine if a document needs an automatic ID applied
         *
         * @param strategy The auto ID strategy for which the document is evaluated
         * @param document The document whose need of an automatic ID should be determined
         * @param idProp The name of the document property containing the ID
         * @return `true` if the document needs an automatic ID, `false` if not
         * @throws DocumentException If bad input prevents the determination
         */
        @Throws(DocumentException::class)
        @JvmStatic
        fun <T> needsAutoId(strategy: AutoId, document: T, idProp: String): Boolean {
            if (document == null) throw DocumentException("document cannot be null")

            if (strategy == DISABLED) return false

            val id = document!!::class.memberProperties.find { it.name == idProp }?.apply { isAccessible = true }
            if (id == null) throw DocumentException("$idProp not found in document")

            if (strategy == NUMBER) {
                return when (id.returnType) {
                    Byte::class.createType()  -> id.call(document) == 0.toByte()
                    Short::class.createType() -> id.call(document) == 0.toShort()
                    Int::class.createType()   -> id.call(document) == 0
                    Long::class.createType()  -> id.call(document) == 0.toLong()
                    else -> throw DocumentException("$idProp was not a number; cannot auto-generate number ID")
                }
            }

            val typ = id.returnType.toString()
            if (typ.endsWith("String") || typ.endsWith("String!")) {
                return id.call(document) == ""
            }

            throw DocumentException("$idProp was not a string ($typ); cannot auto-generate UUID or random string")
        }
    }
}
