package solutions.bitbadger.documents

import java.sql.PreparedStatement
import java.sql.Types
import kotlin.jvm.Throws

/**
 * A parameter to use for a query
 *
 * @property name The name of the parameter (prefixed with a colon)
 * @property type The type of this parameter
 * @property value The value of the parameter
 */
class Parameter<T>(val name: String, val type: ParameterType, val value: T) {

    init {
        if (!name.startsWith(':') && !name.startsWith('@'))
            throw DocumentException("Name must start with : or @ ($name)")
    }

    /**
     * Bind this parameter to a prepared statement at the given index
     *
     * @param stmt The prepared statement to which this parameter should be bound
     * @param index The index (1-based) to which the parameter should be bound
     * @throws DocumentException If a number parameter is given a non-numeric value
     */
    @Throws(DocumentException::class)
    fun bind(stmt: PreparedStatement, index: Int) {
        when (type) {
            ParameterType.NUMBER -> {
                when (value) {
                    null     -> stmt.setNull(index, Types.NULL)
                    is Byte  -> stmt.setByte(index, value)
                    is Short -> stmt.setShort(index, value)
                    is Int   -> stmt.setInt(index, value)
                    is Long  -> stmt.setLong(index, value)
                    else     -> throw DocumentException(
                        "Number parameter must be Byte, Short, Int, or Long (${value::class.simpleName})"
                    )
                }
            }

            ParameterType.STRING -> {
                when (value) {
                    null      -> stmt.setNull(index, Types.NULL)
                    is String -> stmt.setString(index, value)
                    else      -> stmt.setString(index, value.toString())
                }
            }

            ParameterType.JSON -> stmt.setObject(index, value as String, Types.OTHER)
        }
    }

    override fun toString() =
        "$type[$name] = $value"
}
