package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.query.DeleteQuery
import java.sql.Connection
import kotlin.jvm.Throws

/**
 * Functions to delete documents
 */
object Delete {

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId The ID of the document to be deleted
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If no dialect has been configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey> byId(tableName: String, docId: TKey, conn: Connection) =
        Custom.nonQuery(
            DeleteQuery.byId(tableName, docId),
            Parameters.addFields(listOf(Field.equal(Configuration.idField, docId, ":id"))),
            conn
        )

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId The ID of the document to be deleted
     * @throws DocumentException If no connection string has been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey> byId(tableName: String, docId: TKey) =
        Configuration.dbConn().use { byId(tableName, docId, it) }

    /**
     * Delete documents using a field comparison
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null, conn: Connection) {
        val named = Parameters.nameFields(fields)
        Custom.nonQuery(DeleteQuery.byFields(tableName, named, howMatched), Parameters.addFields(named), conn)
    }

    /**
     * Delete documents using a field comparison
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        Configuration.dbConn().use { byFields(tableName, fields, howMatched, it) }

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> byContains(tableName: String, criteria: TContains, conn: Connection) =
        Custom.nonQuery(DeleteQuery.byContains(tableName), listOf(Parameters.json(":criteria", criteria)), conn)

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria The object for which JSON containment should be checked
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> byContains(tableName: String, criteria: TContains) =
        Configuration.dbConn().use { byContains(tableName, criteria, it) }

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path The JSON path comparison to match
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String, path: String, conn: Connection) =
        Custom.nonQuery(DeleteQuery.byJsonPath(tableName), listOf(Parameter(":path", ParameterType.STRING, path)), conn)

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path The JSON path comparison to match
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String, path: String) =
        Configuration.dbConn().use { byJsonPath(tableName, path, it) }
}
