package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.AutoId
import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.query.DocumentQuery
import solutions.bitbadger.documents.query.Where
import solutions.bitbadger.documents.query.statementWhere
import java.sql.Connection
import kotlin.jvm.Throws

/**
 * Functions for manipulating documents
 */
object Document {

    /**
     * Insert a new document
     *
     * @param tableName The table into which the document should be inserted (may include schema)
     * @param document The document to be inserted
     * @param conn The connection on which the query should be executed
     * @throws DocumentException If IDs are misconfigured, or if the database command fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> insert(tableName: String, document: TDoc, conn: Connection) {
        val strategy = Configuration.autoIdStrategy
        val query = if (strategy == AutoId.DISABLED || !AutoId.needsAutoId(strategy, document, Configuration.idField)) {
            DocumentQuery.insert(tableName)
        } else {
            DocumentQuery.insert(tableName, strategy)
        }
        Custom.nonQuery(query, listOf(Parameters.json(":data", document)), conn)
    }

    /**
     * Insert a new document
     *
     * @param tableName The table into which the document should be inserted (may include schema)
     * @param document The document to be inserted
     * @throws DocumentException If no connection string has been set; if IDs are misconfigured; or if the database
     *                           command fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> insert(tableName: String, document: TDoc) =
        Configuration.dbConn().use { insert(tableName, document, it) }

    /**
     * Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table in which the document should be saved (may include schema)
     * @param document The document to be saved
     * @param conn The connection on which the query should be executed
     * @throws DocumentException If the database command fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> save(tableName: String, document: TDoc, conn: Connection) =
        Custom.nonQuery(DocumentQuery.save(tableName), listOf(Parameters.json(":data", document)), conn)

    /**
     * Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table in which the document should be saved (may include schema)
     * @param document The document to be saved
     * @throws DocumentException If no connection string has been set, or if the database command fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> save(tableName: String, document: TDoc) =
        Configuration.dbConn().use { save(tableName, document, it) }

    /**
     * Update (replace) a document by its ID
     *
     * @param tableName The table in which the document should be replaced (may include schema)
     * @param docId The ID of the document to be replaced
     * @param document The document to be replaced
     * @param conn The connection on which the query should be executed
     * @throws DocumentException If no dialect has been configured, or if the database command fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey, TDoc> update(tableName: String, docId: TKey, document: TDoc, conn: Connection) =
        Custom.nonQuery(
            statementWhere(DocumentQuery.update(tableName), Where.byId(":id", docId)),
            Parameters.addFields(
                listOf(Field.equal(Configuration.idField, docId, ":id")),
                mutableListOf(Parameters.json(":data", document))
            ),
            conn
        )

    /**
     * Update (replace) a document by its ID
     *
     * @param tableName The table in which the document should be replaced (may include schema)
     * @param docId The ID of the document to be replaced
     * @param document The document to be replaced
     * @throws DocumentException If no connection string has been set, or if the database command fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey, TDoc> update(tableName: String, docId: TKey, document: TDoc) =
        Configuration.dbConn().use { update(tableName, docId, document, it) }
}
