package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.DocumentException
import java.io.PrintWriter
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.sql.SQLException

/**
 * Functions to create results from queries
 */
object Results {

    /**
     * Create a domain item from a document, specifying the field in which the document is found
     *
     * @param field The field name containing the JSON document
     * @param rs A `ResultSet` set to the row with the document to be constructed
     * @param clazz The class of the document to be returned
     * @return The constructed domain item
     */
    @JvmStatic
    fun <TDoc> fromDocument(field: String, rs: ResultSet, clazz: Class<TDoc>) =
        DocumentConfig.serializer.deserialize(rs.getString(field), clazz)

    /**
     * Create a domain item from a document
     *
     * @param rs A `ResultSet` set to the row with the document to be constructed<
     * @param clazz The class of the document to be returned
     * @return The constructed domain item
     */
    @JvmStatic
    fun <TDoc> fromData(rs: ResultSet, clazz: Class<TDoc>) =
        fromDocument("data", rs, clazz)

    /**
     * Create a list of items for the results of the given command, using the specified mapping function
     *
     * @param stmt The prepared statement to execute
     * @param mapFunc The mapping function from data reader to domain class instance
     * @param clazz The class of the document to be returned
     * @return A list of items from the query's result
     * @throws DocumentException If there is a problem executing the query (unchecked)
     */
    @JvmStatic
    fun <TDoc> toCustomList(
        stmt: PreparedStatement, clazz: Class<TDoc>, mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) =
        try {
            stmt.executeQuery().use {
                val results = mutableListOf<TDoc>()
                while (it.next()) {
                    results.add(mapFunc(it, clazz))
                }
                results.toList()
            }
        } catch (ex: SQLException) {
            throw DocumentException("Error retrieving documents from query: ${ex.message}", ex)
        }

    /**
     * Extract a count from the first column
     *
     * @param rs A `ResultSet` set to the row with the count to retrieve
     * @param clazz The type parameter (ignored; this always returns `Long`)
     * @return The count from the row
     * @throws DocumentException If the dialect has not been set (unchecked)
     */
    @JvmStatic
    fun toCount(rs: ResultSet, clazz: Class<*>) =
        when (Configuration.dialect()) {
            Dialect.POSTGRESQL -> rs.getInt("it").toLong()
            Dialect.SQLITE -> rs.getLong("it")
        }

    /**
     * Extract a true/false value from the first column
     *
     * @param rs A `ResultSet` set to the row with the true/false value to retrieve
     * @param clazz The type parameter (ignored; this always returns `Boolean`)
     * @return The true/false value from the row
     * @throws DocumentException If the dialect has not been set (unchecked)
     */
    @JvmStatic
    fun toExists(rs: ResultSet, clazz: Class<*>) =
        when (Configuration.dialect()) {
            Dialect.POSTGRESQL -> rs.getBoolean("it")
            Dialect.SQLITE -> toCount(rs, Long::class.java) > 0L
        }

    /**
     * Retrieve the JSON text of a document, specifying the field in which the document is found
     *
     * @param field The field name containing the JSON document
     * @param rs A `ResultSet` set to the row with the document to be constructed
     * @return The JSON text of the document
     */
    @JvmStatic
    fun jsonFromDocument(field: String, rs: ResultSet) =
        rs.getString(field) ?: "{}"

    /**
     * Retrieve the JSON text of a document, specifying the field in which the document is found
     *
     * @param rs A `ResultSet` set to the row with the document to be constructed
     * @return The JSON text of the document
     */
    @JvmStatic
    fun jsonFromData(rs: ResultSet) =
        jsonFromDocument("data", rs)

    /**
     * Create a JSON array of items for the results of the given command, using the specified mapping function
     *
     * @param stmt The prepared statement to execute
     * @param mapFunc The mapping function from data reader to JSON text
     * @return A string with a JSON array of documents from the query's result
     * @throws DocumentException If there is a problem executing the query (unchecked)
     */
    @JvmStatic
    fun toJsonArray(stmt: PreparedStatement, mapFunc: (ResultSet) -> String): String =
        try {
            val results = StringBuilder("[")
            stmt.executeQuery().use {
                while (it.next()) {
                    if (results.length > 2) results.append(",")
                    results.append(mapFunc(it))
                }
            }
            results.append("]").toString()
        } catch (ex: SQLException) {
            throw DocumentException("Error retrieving documents from query: ${ex.message}", ex)
        }

    /**
     * Write a JSON array of items for the results of the given command to the given `PrintWriter`, using the specified
     * mapping function
     *
     * @param writer The writer for the results of the query
     * @param stmt The prepared statement to execute
     * @param mapFunc The mapping function from data reader to JSON text
     * @return A string with a JSON array of documents from the query's result
     * @throws DocumentException If there is a problem executing the query (unchecked)
     */
    @JvmStatic
    fun writeJsonArray(writer: PrintWriter, stmt: PreparedStatement, mapFunc: (ResultSet) -> String) =
        try {
            writer.write("[")
            stmt.executeQuery().use {
                var isFirst = true
                while (it.next()) {
                    if (isFirst) {
                        isFirst = false
                    } else {
                        writer.write(",")
                    }
                    writer.write(mapFunc(it))
                }
            }
            writer.write("]")
        } catch (ex: SQLException) {
            throw DocumentException("Error writing documents from query: ${ex.message}", ex)
        }
}
