package solutions.bitbadger.documents.core.tests.java;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.Dialect;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.FieldMatch;
import solutions.bitbadger.documents.query.QueryUtils;
import solutions.bitbadger.documents.core.tests.ForceDialect;

import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;

/**
 * Unit tests for the package-level query functions, presented as `QueryUtils` for Java-compatible use
 */
@DisplayName("Core | Java | Query | QueryUtils")
final public class QueryUtilsTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    public void cleanUp() {
        ForceDialect.none();
    }

    @Test
    @DisplayName("statementWhere generates correctly")
    public void statementWhere() {
        assertEquals("x WHERE y", QueryUtils.statementWhere("x", "y"), "Statements not combined correctly");
    }

    @Test
    @DisplayName("byId generates a numeric ID query | PostgreSQL")
    public void byIdNumericPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("test WHERE (data->>'id')::numeric = :id", QueryUtils.byId("test", 9));
    }

    @Test
    @DisplayName("byId generates an alphanumeric ID query | PostgreSQL")
    public void byIdAlphaPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("unit WHERE data->>'id' = :id", QueryUtils.byId("unit", "18"));
    }

    @Test
    @DisplayName("byId generates ID query | SQLite")
    public void byIdSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("yo WHERE data->>'id' = :id", QueryUtils.byId("yo", 27));
    }

    @Test
    @DisplayName("byFields generates default field query | PostgreSQL")
    public void byFieldsMultipleDefaultPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value",
                QueryUtils.byFields("this", List.of(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value"))));
    }

    @Test
    @DisplayName("byFields generates default field query | SQLite")
    public void byFieldsMultipleDefaultSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("this WHERE data->>'a' = :the_a AND data->>'b' = :b_value",
                QueryUtils.byFields("this", List.of(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value"))));
    }

    @Test
    @DisplayName("byFields generates ANY field query | PostgreSQL")
    public void byFieldsMultipleAnyPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value",
                QueryUtils.byFields("that", List.of(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                        FieldMatch.ANY));
    }

    @Test
    @DisplayName("byFields generates ANY field query | SQLite")
    public void byFieldsMultipleAnySQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("that WHERE data->>'a' = :the_a OR data->>'b' = :b_value",
                QueryUtils.byFields("that", List.of(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                        FieldMatch.ANY));
    }

    @Test
    @DisplayName("orderBy generates for no fields")
    public void orderByNone() throws DocumentException {
        assertEquals("", QueryUtils.orderBy(List.of(), Dialect.POSTGRESQL),
                "ORDER BY should have been blank (PostgreSQL)");
        assertEquals("", QueryUtils.orderBy(List.of(), Dialect.SQLITE), "ORDER BY should have been blank (SQLite)");
    }

    @Test
    @DisplayName("orderBy generates single, no direction | PostgreSQL")
    public void orderBySinglePostgres() throws DocumentException {
        assertEquals(" ORDER BY data->>'TestField'",
                QueryUtils.orderBy(List.of(Field.named("TestField")), Dialect.POSTGRESQL),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates single, no direction | SQLite")
    public void orderBySingleSQLite() throws DocumentException {
        assertEquals(" ORDER BY data->>'TestField'",
                QueryUtils.orderBy(List.of(Field.named("TestField")), Dialect.SQLITE),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates multiple with direction | PostgreSQL")
    public void orderByMultiplePostgres() throws DocumentException {
        assertEquals(" ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
                QueryUtils.orderBy(List.of(
                        Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                        Dialect.POSTGRESQL),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates multiple with direction | SQLite")
    public void orderByMultipleSQLite() throws DocumentException {
        assertEquals(" ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
                QueryUtils.orderBy(List.of(
                        Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                        Dialect.SQLITE),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates numeric ordering | PostgreSQL")
    public void orderByNumericPostgres() throws DocumentException {
        assertEquals(" ORDER BY (data->>'Test')::numeric",
                QueryUtils.orderBy(List.of(Field.named("n:Test")), Dialect.POSTGRESQL),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates numeric ordering | SQLite")
    public void orderByNumericSQLite() throws DocumentException {
        assertEquals(" ORDER BY data->>'Test'", QueryUtils.orderBy(List.of(Field.named("n:Test")), Dialect.SQLITE),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates case-insensitive ordering | PostgreSQL")
    public void orderByCIPostgres() throws DocumentException {
        assertEquals(" ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
                QueryUtils.orderBy(List.of(Field.named("i:Test.Field DESC NULLS FIRST")), Dialect.POSTGRESQL),
                "ORDER BY not constructed correctly");
    }

    @Test
    @DisplayName("orderBy generates case-insensitive ordering | SQLite")
    public void orderByCISQLite() throws DocumentException {
        assertEquals(" ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
                QueryUtils.orderBy(List.of(Field.named("i:Test.Field ASC NULLS LAST")), Dialect.SQLITE),
                "ORDER BY not constructed correctly");
    }
}
