package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.AutoId;
import solutions.bitbadger.documents.Configuration;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;

import java.util.List;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;

final public class DocumentFunctions {

    public static void insertDefault(ThrowawayDatabase db) throws DocumentException {
        assertEquals(0L, countAll(db.getConn(), TEST_TABLE), "There should be no documents in the table");
        insert(db.getConn(), TEST_TABLE, new JsonDocument("turkey", "", 0, new SubDocument("gobble", "gobble!")));
        final List<JsonDocument> after = findAll(db.getConn(), TEST_TABLE, JsonDocument.class);
        assertEquals(1, after.size(), "There should be one document in the table");
        final JsonDocument doc = after.get(0);
        assertEquals("turkey", doc.getId(), "The inserted document's ID is incorrect");
        assertEquals("", doc.getValue(), "The inserted document's value is incorrect");
        assertEquals(0, doc.getNumValue(), "The document's numeric value is incorrect");
        assertNotNull(doc.getSub(), "The inserted document's subdocument was not created");
        assertEquals("gobble", doc.getSub().getFoo(), "The subdocument's \"foo\" property is incorrect");
        assertEquals("gobble!", doc.getSub().getBar(), "The subdocument's \"bar\" property is incorrect");
    }

    public static void insertDupe(ThrowawayDatabase db) throws DocumentException {
        insert(db.getConn(), TEST_TABLE, new JsonDocument("a", "", 0));
        assertThrows(DocumentException.class, () -> insert(db.getConn(), TEST_TABLE, new JsonDocument("a", "b", 22)),
                "Inserting a document with a duplicate key should have thrown an exception");
    }

    public static void insertNumAutoId(ThrowawayDatabase db) throws DocumentException {
        try {
            Configuration.autoIdStrategy = AutoId.NUMBER;
            Configuration.idField = "key";
            assertEquals(0L, countAll(db.getConn(), TEST_TABLE), "There should be no documents in the table");

            insert(db.getConn(), TEST_TABLE, new NumIdDocument(0, "one"));
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(0, "two"));
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(77, "three"));
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(0, "four"));

            final List<NumIdDocument> after = findAll(db.getConn(), TEST_TABLE, NumIdDocument.class,
                    List.of(Field.named("key")));
            assertEquals(4, after.size(), "There should have been 4 documents returned");
            assertEquals("1|2|77|78",
                    after.stream().map(doc -> String.valueOf(doc.getKey()))
                            .reduce((acc, item) -> String.format("%s|%s", acc, item)).get(),
                    "The IDs were not generated correctly");
        } finally {
            Configuration.autoIdStrategy = AutoId.DISABLED;
            Configuration.idField = "id";
        }
    }

    public static void insertUUIDAutoId(ThrowawayDatabase db) throws DocumentException {
        try {
            Configuration.autoIdStrategy = AutoId.UUID;
            assertEquals(0L, countAll(db.getConn(), TEST_TABLE), "There should be no documents in the table");

            insert(db.getConn(), TEST_TABLE, new JsonDocument(""));

            final List<JsonDocument> after = findAll(db.getConn(), TEST_TABLE, JsonDocument.class);
            assertEquals(1, after.size(), "There should have been 1 document returned");
            assertEquals(32, after.get(0).getId().length(), "The ID was not generated correctly");
        } finally {
            Configuration.autoIdStrategy = AutoId.DISABLED;
        }
    }

    public static void insertStringAutoId(ThrowawayDatabase db) throws DocumentException {
        try {
            Configuration.autoIdStrategy = AutoId.RANDOM_STRING;
            assertEquals(0L, countAll(db.getConn(), TEST_TABLE), "There should be no documents in the table");

            insert(db.getConn(), TEST_TABLE, new JsonDocument(""));

            Configuration.idStringLength = 21;
            insert(db.getConn(), TEST_TABLE, new JsonDocument(""));

            final List<JsonDocument> after = findAll(db.getConn(), TEST_TABLE, JsonDocument.class);
            assertEquals(2, after.size(), "There should have been 2 documents returned");
            assertEquals(16, after.get(0).getId().length(), "The first document's ID was not generated correctly");
            assertEquals(21, after.get(1).getId().length(), "The second document's ID was not generated correctly");
        } finally {
            Configuration.autoIdStrategy = AutoId.DISABLED;
            Configuration.idStringLength = 16;
        }
    }

    public static void saveMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        save(db.getConn(), TEST_TABLE, new JsonDocument("two", "", 44));
        final Optional<JsonDocument> tryDoc = findById(db.getConn(), TEST_TABLE, "two", JsonDocument.class);
        assertTrue(tryDoc.isPresent(), "There should have been a document returned");
        final JsonDocument doc = tryDoc.get();
        assertEquals("two", doc.getId(), "An incorrect document was returned");
        assertEquals("", doc.getValue(), "The \"value\" field was not updated");
        assertEquals(44, doc.getNumValue(), "The \"numValue\" field was not updated");
        assertNull(doc.getSub(), "The \"sub\" field was not updated");
    }

    public static void saveNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final JsonDocument toSave = new JsonDocument("test");
        toSave.setSub(new SubDocument("a", "b"));
        save(db.getConn(), TEST_TABLE, toSave);
        assertTrue(findById(db.getConn(), TEST_TABLE, "test", JsonDocument.class).isPresent(),
                "The test document should have been saved");
    }

    public static void updateMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        update(db.getConn(), TEST_TABLE, "one", new JsonDocument("one", "howdy", 8, new SubDocument("y", "z")));
        final Optional<JsonDocument> tryDoc = findById(db.getConn(), TEST_TABLE, "one", JsonDocument.class);
        assertTrue(tryDoc.isPresent(), "There should have been a document returned");
        final JsonDocument doc = tryDoc.get();
        assertEquals("one", doc.getId(), "An incorrect document was returned");
        assertEquals("howdy", doc.getValue(), "The \"value\" field was not updated");
        assertEquals(8, doc.getNumValue(), "The \"numValue\" field was not updated");
        assertNotNull(doc.getSub(), "The sub-document should not be null");
        assertEquals("y", doc.getSub().getFoo(), "The sub-document \"foo\" field was not updated");
        assertEquals("z", doc.getSub().getBar(), "The sub-document \"bar\" field was not updated");
    }

    public static void updateNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsById(db.getConn(), TEST_TABLE, "two-hundred"));
        update(db.getConn(), TEST_TABLE, "two-hundred", new JsonDocument("two-hundred", "", 200));
        assertFalse(existsById(db.getConn(), TEST_TABLE, "two-hundred"));
    }
}
