package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;

import java.util.List;
import java.util.Map;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;

final public class PatchFunctions {

    public static void byIdMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        patchById(db.getConn(), TEST_TABLE, "one", Map.of("numValue", 44));
        final Optional<JsonDocument> doc = findById(db.getConn(), TEST_TABLE, "one", JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("one", doc.get().getId(), "An incorrect document was returned");
        assertEquals(44, doc.get().getNumValue(), "The document was not patched");
    }

    public static void byIdNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsById(db.getConn(), TEST_TABLE, "forty-seven"),
                "Document with ID \"forty-seven\" should not exist");
        patchById(db.getConn(), TEST_TABLE, "forty-seven", Map.of("foo", "green")); // no exception = pass
    }

    public static void byFieldsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        patchByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "purple")), Map.of("numValue", 77));
        assertEquals(2L, countByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("numValue", 77))),
                "There should have been 2 documents with numeric value 77");
    }

    public static void byFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<Field<?>> fields = List.of(Field.equal("value", "burgundy"));
        assertFalse(existsByFields(db.getConn(), TEST_TABLE, fields),
                "There should be no documents with value of \"burgundy\"");
        patchByFields(db.getConn(), TEST_TABLE, fields, Map.of("foo", "green")); // no exception = pass
    }

    public static void byContainsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Map<String, String> contains = Map.of("value", "another");
        patchByContains(db.getConn(), TEST_TABLE, contains, Map.of("numValue", 12));
        final Optional<JsonDocument> doc = findFirstByContains(db.getConn(), TEST_TABLE, contains, JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("two", doc.get().getId(), "The incorrect document was returned");
        assertEquals(12, doc.get().getNumValue(), "The document was not updated");
    }

    public static void byContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Map<String, String> contains = Map.of("value", "updated");
        assertFalse(existsByContains(db.getConn(), TEST_TABLE, contains), "There should be no matching documents");
        patchByContains(db.getConn(), TEST_TABLE, contains, Map.of("sub.foo", "green")); // no exception = pass
    }

    public static void byJsonPathMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String path = "$.numValue ? (@ > 10)";
        patchByJsonPath(db.getConn(), TEST_TABLE, path, Map.of("value", "blue"));
        final List<JsonDocument> docs = findByJsonPath(db.getConn(), TEST_TABLE, path, JsonDocument.class);
        assertEquals(2, docs.size(), "There should have been two documents returned");
        for (final JsonDocument doc : docs) {
            assertTrue(List.of("four", "five").contains(doc.getId()),
                    String.format("An incorrect document was returned (%s)", doc.getId()));
            assertEquals("blue", doc.getValue(), String.format("The value for ID %s was incorrect", doc.getId()));
        }
    }

    public static void byJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String path = "$.numValue ? (@ > 100)";
        assertFalse(existsByJsonPath(db.getConn(), TEST_TABLE, path),
                "There should be no documents with numeric values over 100");
        patchByJsonPath(db.getConn(), TEST_TABLE, path, Map.of("value", "blue")); // no exception = pass
    }
}
