package solutions.bitbadger.documents.core.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.DocumentIndex
import solutions.bitbadger.documents.query.DefinitionQuery
import kotlin.test.assertEquals

/**
 * Unit tests for the `Definition` object
 */
@DisplayName("Core | Kotlin | Query | DefinitionQuery")
class DefinitionQueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName("ensureTableFor generates correctly")
    fun ensureTableFor() =
        assertEquals(
            "CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)",
            DefinitionQuery.ensureTableFor("my.table", "JSONB"), "CREATE TABLE statement not constructed correctly"
        )

    @Test
    @DisplayName("ensureTable generates correctly | PostgreSQL")
    fun ensureTablePostgres() {
        ForceDialect.postgres()
        assertEquals(
            "CREATE TABLE IF NOT EXISTS $TEST_TABLE (data JSONB NOT NULL)",
            DefinitionQuery.ensureTable(TEST_TABLE)
        )
    }

    @Test
    @DisplayName("ensureTable generates correctly | SQLite")
    fun ensureTableSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "CREATE TABLE IF NOT EXISTS $TEST_TABLE (data TEXT NOT NULL)",
            DefinitionQuery.ensureTable(TEST_TABLE)
        )
    }

    @Test
    @DisplayName("ensureTable fails when no dialect is set")
    fun ensureTableFailsUnknown() {
        assertThrows<DocumentException> { DefinitionQuery.ensureTable(TEST_TABLE) }
    }

    @Test
    @DisplayName("ensureKey generates correctly with schema")
    fun ensureKeyWithSchema() =
        assertEquals(
            "CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
            DefinitionQuery.ensureKey("test.table", Dialect.POSTGRESQL),
            "CREATE INDEX for key statement with schema not constructed correctly"
        )

    @Test
    @DisplayName("ensureKey generates correctly without schema")
    fun ensureKeyWithoutSchema() =
        assertEquals(
            "CREATE UNIQUE INDEX IF NOT EXISTS idx_${TEST_TABLE}_key ON $TEST_TABLE ((data->>'id'))",
            DefinitionQuery.ensureKey(TEST_TABLE, Dialect.SQLITE),
            "CREATE INDEX for key statement without schema not constructed correctly"
        )

    @Test
    @DisplayName("ensureIndexOn generates multiple fields and directions")
    fun ensureIndexOnMultipleFields() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
            DefinitionQuery.ensureIndexOn(
                "test.table", "gibberish", listOf("taco", "guac DESC", "salsa ASC"),
                Dialect.POSTGRESQL
            ),
            "CREATE INDEX for multiple field statement not constructed correctly"
        )

    @Test
    @DisplayName("ensureIndexOn generates nested field | PostgreSQL")
    fun ensureIndexOnNestedPostgres() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data#>>'{a,b,c}'))",
            DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", listOf("a.b.c"), Dialect.POSTGRESQL),
            "CREATE INDEX for nested PostgreSQL field incorrect"
        )

    @Test
    @DisplayName("ensureIndexOn generates nested field | SQLite")
    fun ensureIndexOnNestedSQLite() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data->'a'->'b'->>'c'))",
            DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", listOf("a.b.c"), Dialect.SQLITE),
            "CREATE INDEX for nested SQLite field incorrect"
        )

    @Test
    @DisplayName("ensureDocumentIndexOn generates Full | PostgreSQL")
    fun ensureDocumentIndexOnFullPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data)",
            DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL),
            "CREATE INDEX for full document index incorrect"
        )
    }

    @Test
    @DisplayName("ensureDocumentIndexOn generates Optimized | PostgreSQL")
    fun ensureDocumentIndexOnOptimizedPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data jsonb_path_ops)",
            DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.OPTIMIZED),
            "CREATE INDEX for optimized document index incorrect"
        )
    }

    @Test
    @DisplayName("ensureDocumentIndexOn fails | SQLite")
    fun ensureDocumentIndexOnFailsSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL) }
    }
}
