package solutions.bitbadger.documents.core.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.query.FindQuery
import kotlin.test.assertEquals

/**
 * Unit tests for the `Find` object
 */
@DisplayName("Core | Kotlin | Query | FindQuery")
class FindQueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName("all generates correctly")
    fun all() =
        assertEquals("SELECT data FROM $TEST_TABLE", FindQuery.all(TEST_TABLE), "Find query not constructed correctly")

    @Test
    @DisplayName("byId generates correctly | PostgreSQL")
    fun byIdPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "SELECT data FROM $TEST_TABLE WHERE data->>'id' = :id",
            FindQuery.byId<String>(TEST_TABLE), "Find query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byId generates correctly | SQLite")
    fun byIdSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "SELECT data FROM $TEST_TABLE WHERE data->>'id' = :id",
            FindQuery.byId<String>(TEST_TABLE), "Find query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byFields generates correctly | PostgreSQL")
    fun byFieldsPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "SELECT data FROM $TEST_TABLE WHERE data->>'a' = :b AND (data->>'c')::numeric < :d",
            FindQuery.byFields(TEST_TABLE, listOf(Field.equal("a", "", ":b"), Field.less("c", 14, ":d"))),
            "Find query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byFields generates correctly | SQLite")
    fun byFieldsSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "SELECT data FROM $TEST_TABLE WHERE data->>'a' = :b AND data->>'c' < :d",
            FindQuery.byFields(TEST_TABLE, listOf(Field.equal("a", "", ":b"), Field.less("c", 14, ":d"))),
            "Find query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byContains generates correctly | PostgreSQL")
    fun byContainsPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "SELECT data FROM $TEST_TABLE WHERE data @> :criteria", FindQuery.byContains(TEST_TABLE),
            "Find query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byContains fails | SQLite")
    fun byContainsSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { FindQuery.byContains(TEST_TABLE) }
    }

    @Test
    @DisplayName("byJsonPath generates correctly | PostgreSQL")
    fun byJsonPathPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "SELECT data FROM $TEST_TABLE WHERE jsonb_path_exists(data, :path::jsonpath)",
            FindQuery.byJsonPath(TEST_TABLE),
            "Find query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byJsonPath fails | SQLite")
    fun byJsonPathSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { FindQuery.byJsonPath(TEST_TABLE) }
    }
}
