package solutions.bitbadger.documents.core.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import solutions.bitbadger.documents.query.*
import kotlin.test.assertEquals

/**
 * Unit tests for the package-level query functions
 */
@DisplayName("Core | Kotlin | Query | Package Functions")
class QueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName("statementWhere generates correctly")
    fun statementWhere() =
        assertEquals("x WHERE y", statementWhere("x", "y"), "Statements not combined correctly")

    @Test
    @DisplayName("byId generates a numeric ID query | PostgreSQL")
    fun byIdNumericPostgres() {
        ForceDialect.postgres()
        assertEquals("test WHERE (data->>'id')::numeric = :id", byId("test", 9))
    }

    @Test
    @DisplayName("byId generates an alphanumeric ID query | PostgreSQL")
    fun byIdAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals("unit WHERE data->>'id' = :id", byId("unit", "18"))
    }

    @Test
    @DisplayName("byId generates ID query | SQLite")
    fun byIdSQLite() {
        ForceDialect.sqlite()
        assertEquals("yo WHERE data->>'id' = :id", byId("yo", 27))
    }

    @Test
    @DisplayName("byFields generates default field query | PostgreSQL")
    fun byFieldsMultipleDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value",
            byFields("this", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")))
        )
    }

    @Test
    @DisplayName("byFields generates default field query | SQLite")
    fun byFieldsMultipleDefaultSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "this WHERE data->>'a' = :the_a AND data->>'b' = :b_value",
            byFields("this", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")))
        )
    }

    @Test
    @DisplayName("byFields generates ANY field query | PostgreSQL")
    fun byFieldsMultipleAnyPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value",
            byFields(
                "that", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                FieldMatch.ANY
            )
        )
    }

    @Test
    @DisplayName("byFields generates ANY field query | SQLite")
    fun byFieldsMultipleAnySQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "that WHERE data->>'a' = :the_a OR data->>'b' = :b_value",
            byFields(
                "that", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                FieldMatch.ANY
            )
        )
    }

    @Test
    @DisplayName("orderBy generates for no fields")
    fun orderByNone() {
        assertEquals("", orderBy(listOf(), Dialect.POSTGRESQL), "ORDER BY should have been blank (PostgreSQL)")
        assertEquals("", orderBy(listOf(), Dialect.SQLITE), "ORDER BY should have been blank (SQLite)")
    }

    @Test
    @DisplayName("orderBy generates single, no direction | PostgreSQL")
    fun orderBySinglePostgres() =
        assertEquals(
            " ORDER BY data->>'TestField'",
            orderBy(listOf(Field.named("TestField")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates single, no direction | SQLite")
    fun orderBySingleSQLite() =
        assertEquals(
            " ORDER BY data->>'TestField'", orderBy(listOf(Field.named("TestField")), Dialect.SQLITE),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates multiple with direction | PostgreSQL")
    fun orderByMultiplePostgres() =
        assertEquals(
            " ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
            orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.POSTGRESQL
            ),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates multiple with direction | SQLite")
    fun orderByMultipleSQLite() =
        assertEquals(
            " ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
            orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.SQLITE
            ),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates numeric ordering | PostgreSQL")
    fun orderByNumericPostgres() =
        assertEquals(
            " ORDER BY (data->>'Test')::numeric",
            orderBy(listOf(Field.named("n:Test")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates numeric ordering | SQLite")
    fun orderByNumericSQLite() =
        assertEquals(
            " ORDER BY data->>'Test'", orderBy(listOf(Field.named("n:Test")), Dialect.SQLITE),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates case-insensitive ordering | PostgreSQL")
    fun orderByCIPostgres() =
        assertEquals(
            " ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
            orderBy(listOf(Field.named("i:Test.Field DESC NULLS FIRST")), Dialect.POSTGRESQL),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates case-insensitive ordering | SQLite")
    fun orderByCISQLite() =
        assertEquals(
            " ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
            orderBy(listOf(Field.named("i:Test.Field ASC NULLS LAST")), Dialect.SQLITE),
            "ORDER BY not constructed correctly"
        )
}
