package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.DocumentIndex
import solutions.bitbadger.documents.query.DefinitionQuery

import static Types.TEST_TABLE
import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `Definition` object
 */
@DisplayName('Groovy | Query | DefinitionQuery')
class DefinitionQueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    void cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName('ensureTableFor generates correctly')
    void ensureTableFor() {
        assertEquals('CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)',
                DefinitionQuery.ensureTableFor('my.table', 'JSONB'), 'CREATE TABLE statement not constructed correctly')
    }

    @Test
    @DisplayName('ensureTable generates correctly | PostgreSQL')
    void ensureTablePostgres() {
        ForceDialect.postgres()
        assertEquals("CREATE TABLE IF NOT EXISTS $TEST_TABLE (data JSONB NOT NULL)".toString(),
                DefinitionQuery.ensureTable(TEST_TABLE))
    }

    @Test
    @DisplayName('ensureTable generates correctly | SQLite')
    void ensureTableSQLite() {
        ForceDialect.sqlite()
        assertEquals("CREATE TABLE IF NOT EXISTS $TEST_TABLE (data TEXT NOT NULL)".toString(),
                DefinitionQuery.ensureTable(TEST_TABLE))
    }

    @Test
    @DisplayName('ensureTable fails when no dialect is set')
    void ensureTableFailsUnknown() {
        assertThrows(DocumentException) { DefinitionQuery.ensureTable(TEST_TABLE) }
    }

    @Test
    @DisplayName('ensureKey generates correctly with schema')
    void ensureKeyWithSchema() {
        assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
                DefinitionQuery.ensureKey('test.table', Dialect.POSTGRESQL),
                'CREATE INDEX for key statement with schema not constructed correctly')
    }

    @Test
    @DisplayName('ensureKey generates correctly without schema')
    void ensureKeyWithoutSchema() {
        assertEquals(
                "CREATE UNIQUE INDEX IF NOT EXISTS idx_${TEST_TABLE}_key ON $TEST_TABLE ((data->>'id'))".toString(),
                DefinitionQuery.ensureKey(TEST_TABLE, Dialect.SQLITE),
                'CREATE INDEX for key statement without schema not constructed correctly')
    }

    @Test
    @DisplayName('ensureIndexOn generates multiple fields and directions')
    void ensureIndexOnMultipleFields() {
        assertEquals(
                "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
                DefinitionQuery.ensureIndexOn('test.table', 'gibberish', List.of('taco', 'guac DESC', 'salsa ASC'),
                        Dialect.POSTGRESQL),
                'CREATE INDEX for multiple field statement not constructed correctly')
    }

    @Test
    @DisplayName('ensureIndexOn generates nested field | PostgreSQL')
    void ensureIndexOnNestedPostgres() {
        assertEquals("CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data#>>'{a,b,c}'))".toString(),
                DefinitionQuery.ensureIndexOn(TEST_TABLE, 'nest', List.of('a.b.c'), Dialect.POSTGRESQL),
                'CREATE INDEX for nested PostgreSQL field incorrect')
    }

    @Test
    @DisplayName('ensureIndexOn generates nested field | SQLite')
    void ensureIndexOnNestedSQLite() {
        assertEquals(
                "CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data->'a'->'b'->>'c'))".toString(),
                DefinitionQuery.ensureIndexOn(TEST_TABLE, 'nest', List.of('a.b.c'), Dialect.SQLITE),
                'CREATE INDEX for nested SQLite field incorrect')
    }

    @Test
    @DisplayName('ensureDocumentIndexOn generates Full | PostgreSQL')
    void ensureDocumentIndexOnFullPostgres() {
        ForceDialect.postgres()
        assertEquals("CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data)".toString(),
                DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL),
                'CREATE INDEX for full document index incorrect')
    }

    @Test
    @DisplayName('ensureDocumentIndexOn generates Optimized | PostgreSQL')
    void ensureDocumentIndexOnOptimizedPostgres() {
        ForceDialect.postgres()
        assertEquals(
                "CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data jsonb_path_ops)"
                        .toString(),
                DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.OPTIMIZED),
                'CREATE INDEX for optimized document index incorrect')
    }

    @Test
    @DisplayName('ensureDocumentIndexOn fails | SQLite')
    void ensureDocumentIndexOnFailsSQLite() {
        ForceDialect.sqlite()
        assertThrows(DocumentException) { DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL) }
    }
}
