package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.Parameter
import solutions.bitbadger.documents.ParameterType
import solutions.bitbadger.documents.java.Parameters

import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `Parameters` object
 */
@DisplayName('Groovy | Parameters')
class ParametersTest {

    /**
     * Reset the dialect
     */
    @AfterEach
    void cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName('nameFields works with no changes')
    void nameFieldsNoChange() {
        def fields = List.of(Field.equal('a', '', ':test'), Field.exists('q'), Field.equal('b', '', ':me'))
        def named = Parameters.nameFields(fields).toList()
        assertEquals fields.size(), named.size(), 'There should have been 3 fields in the list'
        assertSame fields[0], named[0], 'The first field should be the same'
        assertSame fields[1], named[1], 'The second field should be the same'
        assertSame fields[2], named[2], 'The third field should be the same'
    }

    @Test
    @DisplayName('nameFields works when changing fields')
    void nameFieldsChange() {
        def fields = List.of(Field.equal('a', ''), Field.equal('e', '', ':hi'), Field.equal('b', ''),
                Field.notExists('z'))
        def named = Parameters.nameFields(fields).toList()
        assertEquals fields.size(), named.size(), 'There should have been 4 fields in the list'
        assertNotSame fields[0], named[0], 'The first field should not be the same'
        assertEquals ':field0', named[0].parameterName, 'First parameter name incorrect'
        assertSame fields[1], named[1], 'The second field should be the same'
        assertNotSame fields[2], named[2], 'The third field should not be the same'
        assertEquals ':field1', named[2].parameterName, 'Third parameter name incorrect'
        assertSame fields[3], named[3], 'The fourth field should be the same'
    }

    @Test
    @DisplayName('replaceNamesInQuery replaces successfully')
    void replaceNamesInQuery() {
        def parameters = List.of(new Parameter(':data', ParameterType.JSON, '{}'),
                new Parameter(':data_ext', ParameterType.STRING, ''))
        def query = 'SELECT data, data_ext FROM tbl WHERE data = :data AND data_ext = :data_ext AND more_data = :data'
        assertEquals('SELECT data, data_ext FROM tbl WHERE data = ? AND data_ext = ? AND more_data = ?',
                Parameters.replaceNamesInQuery(query, parameters), 'Parameters not replaced correctly')
    }

    @Test
    @DisplayName('fieldNames generates a single parameter (PostgreSQL)')
    void fieldNamesSinglePostgres() {
        ForceDialect.postgres()
        def nameParams = Parameters.fieldNames(List.of('test')).toList()
        assertEquals 1, nameParams.size(), 'There should be one name parameter'
        assertEquals ':name', nameParams[0].name, 'The parameter name is incorrect'
        assertEquals ParameterType.STRING, nameParams[0].type, 'The parameter type is incorrect'
        assertEquals '{test}', nameParams[0].value, 'The parameter value is incorrect'
    }

    @Test
    @DisplayName('fieldNames generates multiple parameters (PostgreSQL)')
    void fieldNamesMultiplePostgres() {
        ForceDialect.postgres()
        def nameParams = Parameters.fieldNames(List.of('test', 'this', 'today')).toList()
        assertEquals 1, nameParams.size(), 'There should be one name parameter'
        assertEquals ':name', nameParams[0].name, 'The parameter name is incorrect'
        assertEquals ParameterType.STRING, nameParams[0].type, 'The parameter type is incorrect'
        assertEquals '{test,this,today}', nameParams[0].value, 'The parameter value is incorrect'
    }

    @Test
    @DisplayName('fieldNames generates a single parameter (SQLite)')
    void fieldNamesSingleSQLite() {
        ForceDialect.sqlite()
        def nameParams = Parameters.fieldNames(List.of('test')).toList()
        assertEquals 1, nameParams.size(), 'There should be one name parameter'
        assertEquals ':name0', nameParams[0].name, 'The parameter name is incorrect'
        assertEquals ParameterType.STRING, nameParams[0].type, 'The parameter type is incorrect'
        assertEquals 'test', nameParams[0].value, 'The parameter value is incorrect'
    }

    @Test
    @DisplayName('fieldNames generates multiple parameters (SQLite)')
    void fieldNamesMultipleSQLite() {
        ForceDialect.sqlite()
        def nameParams = Parameters.fieldNames(List.of('test', 'this', 'today')).toList()
        assertEquals 3, nameParams.size(), 'There should be one name parameter'
        assertEquals ':name0', nameParams[0].name, 'The first parameter name is incorrect'
        assertEquals ParameterType.STRING, nameParams[0].type, 'The first parameter type is incorrect'
        assertEquals 'test', nameParams[0].value, 'The first parameter value is incorrect'
        assertEquals ':name1', nameParams[1].name, 'The second parameter name is incorrect'
        assertEquals ParameterType.STRING, nameParams[1].type, 'The second parameter type is incorrect'
        assertEquals 'this', nameParams[1].value, 'The second parameter value is incorrect'
        assertEquals ':name2', nameParams[2].name, 'The third parameter name is incorrect'
        assertEquals ParameterType.STRING, nameParams[2].type, 'The third parameter type is incorrect'
        assertEquals 'today', nameParams[2].value, 'The third parameter value is incorrect'
    }

    @Test
    @DisplayName('fieldNames fails if dialect not set')
    void fieldNamesFails() {
        assertThrows(DocumentException) { Parameters.fieldNames(List.of()) }
    }
}
