package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.Parameter
import solutions.bitbadger.documents.ParameterType
import solutions.bitbadger.documents.java.Results
import solutions.bitbadger.documents.query.CountQuery
import solutions.bitbadger.documents.query.DeleteQuery
import solutions.bitbadger.documents.query.FindQuery
import solutions.bitbadger.documents.query.QueryUtils

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

final class CustomFunctions {

    static void listEmpty(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.deleteByFields TEST_TABLE, List.of(Field.exists(Configuration.idField))
        def result = db.conn.customList FindQuery.all(TEST_TABLE), List.of(), JsonDocument, Results.&fromData
        assertEquals 0, result.size(), 'There should have been no results'
    }

    static void listAll(ThrowawayDatabase db) {
        JsonDocument.load db
        def result = db.conn.customList FindQuery.all(TEST_TABLE), List.of(), JsonDocument, Results.&fromData
        assertEquals 5, result.size(), 'There should have been 5 results'
    }

    static void jsonArrayEmpty(ThrowawayDatabase db) {
        assertEquals(0L, db.conn.countAll(TEST_TABLE), 'The test table should be empty')
        assertEquals('[]', db.conn.customJsonArray(FindQuery.all(TEST_TABLE), List.of(), Results.&jsonFromData),
                'An empty list was not represented correctly');
    }

    static void jsonArraySingle(ThrowawayDatabase db) {
        db.conn.insert(TEST_TABLE, new ArrayDocument("one", List.of("2", "3")))
        assertEquals(JsonFunctions.maybeJsonB('[{"id":"one","values":["2","3"]}]'),
                db.conn.customJsonArray(FindQuery.all(TEST_TABLE), List.of(), Results.&jsonFromData),
                'A single document list was not represented correctly')
    }

    static void jsonArrayMany(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        assertEquals(JsonFunctions.maybeJsonB('[{"id":"first","values":["a","b","c"]},'
                        + '{"id":"second","values":["c","d","e"]},{"id":"third","values":["x","y","z"]}]'),
                db.conn.customJsonArray(FindQuery.all(TEST_TABLE) + QueryUtils.orderBy(List.of(Field.named("id"))),
                        List.of(), Results.&jsonFromData),
                'A multiple document list was not represented correctly')
    }

    static void writeJsonArrayEmpty(ThrowawayDatabase db) {
        assertEquals(0L, db.conn.countAll(TEST_TABLE), 'The test table should be empty')
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE), List.of(), writer, Results.&jsonFromData)
        assertEquals("[]", output.toString(), 'An empty list was not represented correctly')
    }

    static void writeJsonArraySingle(ThrowawayDatabase db) {
        db.conn.insert(TEST_TABLE, new ArrayDocument("one", List.of("2", "3")))
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE), List.of(), writer, Results.&jsonFromData)
        assertEquals(JsonFunctions.maybeJsonB('[{"id":"one","values":["2","3"]}]'), output.toString(),
                'A single document list was not represented correctly')
    }

    static void writeJsonArrayMany(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE) + QueryUtils.orderBy(List.of(Field.named("id"))),
                List.of(), writer, Results.&jsonFromData)
        assertEquals(JsonFunctions.maybeJsonB('[{"id":"first","values":["a","b","c"]},'
                        + '{"id":"second","values":["c","d","e"]},{"id":"third","values":["x","y","z"]}]'),
                output.toString(), "A multiple document list was not represented correctly")
    }

    static void singleNone(ThrowawayDatabase db) {
        assertFalse(db.conn.customSingle(FindQuery.all(TEST_TABLE), List.of(), JsonDocument, Results.&fromData)
                        .isPresent(),
                'There should not have been a document returned')

    }

    static void singleOne(ThrowawayDatabase db) {
        JsonDocument.load db
        assertTrue(db.conn.customSingle(FindQuery.all(TEST_TABLE), List.of(), JsonDocument, Results.&fromData)
                        .isPresent(),
                'There should not have been a document returned')
    }

    static void jsonSingleNone(ThrowawayDatabase db) {
        assertEquals('{}', db.conn.customJsonSingle(FindQuery.all(TEST_TABLE), List.of(), Results.&jsonFromData),
                'An empty document was not represented correctly')
    }

    static void jsonSingleOne(ThrowawayDatabase db) {
        db.conn.insert(TEST_TABLE, new ArrayDocument("me", List.of("myself", "i")))
        assertEquals(JsonFunctions.maybeJsonB('{"id":"me","values":["myself","i"]}'),
                db.conn.customJsonSingle(FindQuery.all(TEST_TABLE), List.of(), Results.&jsonFromData),
                'A single document was not represented correctly');
    }

    static void nonQueryChanges(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals(5L, db.conn.customScalar(CountQuery.all(TEST_TABLE), List.of(), Long, Results.&toCount),
                'There should have been 5 documents in the table')
        db.conn.customNonQuery("DELETE FROM $TEST_TABLE")
        assertEquals(0L, db.conn.customScalar(CountQuery.all(TEST_TABLE), List.of(), Long, Results.&toCount),
                'There should have been no documents in the table')
    }

    static void nonQueryNoChanges(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals(5L, db.conn.customScalar(CountQuery.all(TEST_TABLE), List.of(), Long, Results.&toCount),
                'There should have been 5 documents in the table')
        db.conn.customNonQuery(DeleteQuery.byId(TEST_TABLE, 'eighty-two'),
                List.of(new Parameter(':id', ParameterType.STRING, 'eighty-two')))
        assertEquals(5L, db.conn.customScalar(CountQuery.all(TEST_TABLE), List.of(), Long, Results.&toCount),
                'There should still have been 5 documents in the table')
    }

    static void scalar(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals(3L,
                db.conn.customScalar("SELECT 3 AS it FROM $TEST_TABLE LIMIT 1", List.of(), Long, Results.&toCount),
                'The number 3 should have been returned')
    }
}
