package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

/**
 * Tests for the JSON-returning functions
 *
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
final class JsonFunctions {

    /**
     * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
     * This function will do a crude string replacement to match the target string based on the dialect being tested.
     *
     * @param json The JSON which should be returned
     * @return The actual expected JSON based on the database being tested
     */
    static String maybeJsonB(String json) {
        return switch (Configuration.dialect()) {
            case Dialect.SQLITE -> json
            case Dialect.POSTGRESQL -> json.replace ('":', '": ' ).replace (',"', ', "')
        }
    }

    /**
     * Create a snippet of JSON to find a document ID
     *
     * @param id The ID of the document
     * @return A connection-aware ID to check for presence and positioning
     */
    private static String docId(String id) {
        return maybeJsonB("{\"id\":\"$id\"")
    }

    private static void checkAllDefault(String json) {
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.one), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.two), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.three), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found in JSON ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('one')), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(docId('two')), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(docId('three')), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(docId('four')), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(docId('five')), "Document 'five' not found in JSON ($json)")
                break
        }
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void allDefault(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkAllDefault(db.conn.jsonAll(TEST_TABLE))
    }

    static void writeAllDefault(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonAll(TEST_TABLE, writer)
        checkAllDefault(output.toString())
    }

    private static void checkAllEmpty(String json) {
        assertEquals('[]', json, 'There should have been no documents returned')
    }

    static void allEmpty(ThrowawayDatabase db) {
        checkAllEmpty(db.conn.jsonAll(TEST_TABLE))
    }

    static void writeAllEmpty(ThrowawayDatabase db) {
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonAll(TEST_TABLE, writer)
        checkAllEmpty(output.toString())
    }

    private static void checkByIdString(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.two, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void byIdString(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByIdString(db.conn.jsonById(TEST_TABLE, 'two'))
    }

    static void writeByIdString(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonById(TEST_TABLE, writer, 'two')
        checkByIdString(output.toString())
    }

    private static void checkByIdNumber(String json) {
        assertEquals(maybeJsonB('{"key":18,"text":"howdy"}'), json, 'The document should have been found by numeric ID')
    }

    static void byIdNumber(ThrowawayDatabase db) {
        Configuration.idField = 'key'
        try {
            db.conn.insert(TEST_TABLE, new NumIdDocument(18, 'howdy'))
            checkByIdNumber(db.conn.jsonById(TEST_TABLE, 18))
        } finally {
            Configuration.idField = 'id'
        }
    }

    static void writeByIdNumber(ThrowawayDatabase db) {
        Configuration.idField = 'key'
        try {
            db.conn.insert(TEST_TABLE, new NumIdDocument(18, 'howdy'))
            def output = new StringWriter()
            def writer = new PrintWriter(output)
            db.conn.writeJsonById(TEST_TABLE, writer, 18)
            checkByIdNumber(output.toString())
        } finally {
            Configuration.idField = 'id'
        }
    }

    private static void checkByIdNotFound(String json) {
        assertEquals('{}', json, 'There should have been no document returned')
    }

    static void byIdNotFound(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByIdNotFound(db.conn.jsonById(TEST_TABLE, 'x'))
    }

    static void writeByIdNotFound(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonById(TEST_TABLE, writer, 'x')
        checkByIdNotFound(output.toString())
    }

    private static void checkByFieldsMatch(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.four}]".toString(), json, 'The incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId('four')), "The incorrect document was returned ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byFieldsMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByFieldsMatch(db.conn.jsonByFields(TEST_TABLE, List.of(Field.any('value', List.of('blue', 'purple')),
                Field.exists('sub')), FieldMatch.ALL))
    }

    static void writeByFieldsMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, List.of(Field.any('value', List.of('blue', 'purple')),
                Field.exists('sub')), FieldMatch.ALL)
        checkByFieldsMatch(output.toString())
    }

    private static checkByFieldsMatchOrdered(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.five},${JsonDocument.four}]".toString(), json,
                        'The documents were not ordered correctly')
                break
            case Dialect.POSTGRESQL:
                int fiveIdx = json.indexOf(docId('five'))
                int fourIdx = json.indexOf(docId('four'))
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByFieldsMatchOrdered(db.conn.jsonByFields(TEST_TABLE, List.of(Field.equal('value', 'purple')), null,
                List.of(Field.named('id'))))
    }

    static void writeByFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, List.of(Field.equal('value', 'purple')), null,
                List.of(Field.named('id')))
        checkByFieldsMatchOrdered(output.toString())
    }

    private static void checkByFieldsMatchNumIn(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.three}]".toString(), json, 'The incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId('three')), "The incorrect document was returned ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byFieldsMatchNumIn(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByFieldsMatchNumIn(db.conn.jsonByFields(TEST_TABLE, List.of(Field.any('numValue', List.of(2, 4, 6, 8)))))
    }

    static void writeByFieldsMatchNumIn(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, List.of(Field.any('numValue', List.of(2, 4, 6, 8))))
        checkByFieldsMatchNumIn(output.toString())
    }

    private static void checkByFieldsNoMatch(String json) {
        assertEquals('[]', json, 'There should have been no documents returned')
    }

    static void byFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByFieldsNoMatch(db.conn.jsonByFields(TEST_TABLE, List.of(Field.greater('numValue', 100))))
    }

    static void writeByFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, List.of(Field.greater('numValue', 100)))
        checkByFieldsNoMatch(output.toString())
    }

    private static void checkByFieldsMatchInArray(String json) {
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        assertTrue(json.contains(docId('first')), "The 'first' document was not found ($json)")
        assertTrue(json.contains(docId('second')), "The 'second' document was not found ($json)")
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void byFieldsMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        checkByFieldsMatchInArray(db.conn.jsonByFields(TEST_TABLE,
                List.of(Field.inArray('values', TEST_TABLE, List.of('c')))))
    }

    static void writeByFieldsMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, List.of(Field.inArray('values', TEST_TABLE, List.of('c'))))
        checkByFieldsMatchInArray(output.toString())
    }

    private static void checkByFieldsNoMatchInArray(String json) {
        assertEquals('[]', json, 'There should have been no documents returned')
    }

    static void byFieldsNoMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        checkByFieldsNoMatchInArray(db.conn.jsonByFields(TEST_TABLE,
                List.of(Field.inArray('values', TEST_TABLE, List.of('j')))))
    }

    static void writeByFieldsNoMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, List.of(Field.inArray('values', TEST_TABLE, List.of('j'))))
        checkByFieldsNoMatchInArray(output.toString())
    }

    private static void checkByContainsMatch(String json) {
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId('five')), "Document 'five' not found ($json)")
                break
        }
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void byContainsMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByContainsMatch(db.conn.jsonByContains(TEST_TABLE, Map.of('value', 'purple')))
    }

    static void writeByContainsMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByContains(TEST_TABLE, writer, Map.of('value', 'purple'))
        checkByContainsMatch(output.toString())
    }

    private static void checkByContainsMatchOrdered(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.two},${JsonDocument.four}]", json,
                        'The documents were not ordered correctly')
                break
            case Dialect.POSTGRESQL:
                int twoIdx = json.indexOf(docId('two'))
                int fourIdx = json.indexOf(docId('four'))
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(twoIdx >= 0, "Document 'two' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(twoIdx < fourIdx, "Document 'two' should have been before 'four' ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByContainsMatchOrdered(db.conn.jsonByContains(TEST_TABLE, Map.of('sub', Map.of('foo', 'green')),
                List.of(Field.named('value'))))
    }

    static void writeByContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByContains(TEST_TABLE, writer, Map.of('sub', Map.of('foo', 'green')),
                List.of(Field.named('value')))
        checkByContainsMatchOrdered(output.toString())
    }

    private static void checkByContainsNoMatch(String json) {
        assertEquals('[]', json, 'There should have been no documents returned')
    }

    static void byContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByContainsNoMatch(db.conn.jsonByContains(TEST_TABLE, Map.of('value', 'indigo')))
    }

    static void writeByContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByContains(TEST_TABLE, writer, Map.of('value', 'indigo'))
        checkByContainsNoMatch(output.toString())
    }

    private static void checkByJsonPathMatch(String json) {
        assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId('five')), "Document 'five' not found ($json)")
                break
        }
        assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
    }

    static void byJsonPathMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByJsonPathMatch(db.conn.jsonByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)'))
    }

    static void writeByJsonPathMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ > 10)')
        checkByJsonPathMatch(output.toString())
    }

    private static void checkByJsonPathMatchOrdered(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals("[${JsonDocument.five},${JsonDocument.four}]", json,
                        'The documents were not ordered correctly')
                break
            case Dialect.POSTGRESQL:
                int fiveIdx = json.indexOf(docId('five'))
                int fourIdx = json.indexOf(docId('four'))
                assertTrue(json.startsWith('['), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith(']'), "JSON should end with ']' ($json)")
                break
        }
    }

    static void byJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByJsonPathMatchOrdered(db.conn.jsonByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)',
                List.of(Field.named('id'))))
    }

    static void writeByJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ > 10)', List.of(Field.named('id')))
        checkByJsonPathMatchOrdered(output.toString())
    }

    private static void checkByJsonPathNoMatch(String json) {
        assertEquals('[]', json, 'There should have been no documents returned')
    }

    static void byJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkByJsonPathNoMatch(db.conn.jsonByJsonPath(TEST_TABLE, '$.numValue ? (@ > 100)'))
    }

    static void writeByJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ > 100)')
        checkByJsonPathNoMatch(output.toString())
    }

    private static void checkFirstByFieldsMatchOne(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.two, json, 'The incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')), "The incorrect document was returned ($json)")
                break
        }
    }

    static void firstByFieldsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByFieldsMatchOne(db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('value', 'another'))))
    }

    static void writeFirstByFieldsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, List.of(Field.equal('value', 'another')))
        checkFirstByFieldsMatchOne(output.toString())
    }

    private static void checkFirstByFieldsMatchMany(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
                        "Expected document 'two' or 'four' ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')) || json.contains(docId('four')),
                        "Expected document 'two' or 'four' ($json)")
                break
        }
    }

    static void firstByFieldsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByFieldsMatchMany(db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('sub.foo', 'green'))))
    }

    static void writeFirstByFieldsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, List.of(Field.equal('sub.foo', 'green')))
        checkFirstByFieldsMatchMany(output.toString())
    }

    private static void checkFirstByFieldsMatchOrdered(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.four, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByFieldsMatchOrdered(db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('sub.foo', 'green')),
                null, List.of(Field.named('n:numValue DESC'))))
    }

    static void writeFirstByFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, List.of(Field.equal('sub.foo', 'green')),
                null, List.of(Field.named('n:numValue DESC')))
        checkFirstByFieldsMatchOrdered(output.toString())
    }

    private static void checkFirstByFieldsNoMatch(String json) {
        assertEquals('{}', json, 'There should have been no document returned')
    }

    static void firstByFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByFieldsNoMatch(db.conn.jsonFirstByFields(TEST_TABLE, List.of(Field.equal('value', 'absent'))))
    }

    static void writeFirstByFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, List.of(Field.equal('value', 'absent')))
        checkFirstByFieldsNoMatch(output.toString())
    }

    private static void checkFirstByContainsMatchOne(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.one, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('one')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByContainsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByContainsMatchOne(db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'FIRST!')))
    }

    static void writeFirstByContainsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.of('value', 'FIRST!'))
        checkFirstByContainsMatchOne(output.toString())
    }

    private static void checkFirstByContainsMatchMany(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                        "Expected document 'four' or 'five' ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue( json.contains(docId('four')) || json.contains(docId('five')),
                        "Expected document 'four' or 'five' ($json)")
                break
        }
    }

    static void firstByContainsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByContainsMatchMany(db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'purple')))
    }

    static void writeFirstByContainsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.of('value', 'purple'))
        checkFirstByContainsMatchMany(output.toString())
    }

    private static void checkFirstByContainsMatchOrdered(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.five, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('five')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByContainsMatchOrdered(db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'purple'),
                List.of(Field.named('sub.bar NULLS FIRST'))))
    }

    static void writeFirstByContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.of('value', 'purple'),
                List.of(Field.named('sub.bar NULLS FIRST')))
        checkFirstByContainsMatchOrdered(output.toString())
    }

    private static void checkFirstByContainsNoMatch(String json) {
        assertEquals('{}', json, 'There should have been no document returned')
    }

    static void firstByContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByContainsNoMatch(db.conn.jsonFirstByContains(TEST_TABLE, Map.of('value', 'indigo')))
    }

    static void writeFirstByContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.of('value', 'indigo'))
        checkFirstByContainsNoMatch(output.toString())
    }

    private static void checkFirstByJsonPathMatchOne(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.two, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('two')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByJsonPathMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByJsonPathMatchOne(db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ == 10)'))
    }

    static void writeFirstByJsonPathMatchOne(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ == 10)')
        checkFirstByJsonPathMatchOne(output.toString())
    }

    private static void checkFirstByJsonPathMatchMany(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                        "Expected document 'four' or 'five' ($json)")
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')) || json.contains(docId('five')),
                        "Expected document 'four' or 'five' ($json)")
                break
        }
    }

    static void firstByJsonPathMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByJsonPathMatchMany(db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)'))
    }

    static void writeFirstByJsonPathMatchMany(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ > 10)')
        checkFirstByJsonPathMatchMany(output.toString())
    }

    private static void checkFirstByJsonPathMatchOrdered(String json) {
        switch (Configuration.dialect()) {
            case Dialect.SQLITE:
                assertEquals(JsonDocument.four, json, 'An incorrect document was returned')
                break
            case Dialect.POSTGRESQL:
                assertTrue(json.contains(docId('four')), "An incorrect document was returned ($json)")
                break
        }
    }

    static void firstByJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByJsonPathMatchOrdered(db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)',
                List.of(Field.named('id DESC'))))
    }

    static void writeFirstByJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ > 10)', List.of(Field.named('id DESC')))
        checkFirstByJsonPathMatchOrdered(output.toString())
    }

    private static void checkFirstByJsonPathNoMatch(String json) {
        assertEquals('{}', json, 'There should have been no document returned')
    }

    static void firstByJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        checkFirstByJsonPathNoMatch(db.conn.jsonFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 100)'))
    }

    static void writeFirstByJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load(db)
        def output = new StringWriter()
        def writer = new PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, '$.numValue ? (@ > 100)')
        checkFirstByJsonPathNoMatch(output.toString())
    }
}
