package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.Field

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

final class PatchFunctions {

    static void byIdMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.patchById TEST_TABLE, 'one', Map.of('numValue', 44)
        def doc = db.conn.findById TEST_TABLE, 'one', JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'one', doc.get().id, 'An incorrect document was returned'
        assertEquals 44, doc.get().numValue, 'The document was not patched'
    }

    static void byIdNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsById(TEST_TABLE, 'forty-seven'), 'Document with ID "forty-seven" should not exist'
        db.conn.patchById TEST_TABLE, 'forty-seven', Map.of('foo', 'green') // no exception = pass
    }

    static void byFieldsMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.patchByFields TEST_TABLE, List.of(Field.equal('value', 'purple')), Map.of('numValue', 77)
        assertEquals(2, db.conn.countByFields(TEST_TABLE, List.of(Field.equal('numValue', 77))),
                'There should have been 2 documents with numeric value 77')
    }

    static void byFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def fields = List.of Field.equal('value', 'burgundy')
        assertFalse db.conn.existsByFields(TEST_TABLE, fields), 'There should be no documents with value of "burgundy"'
        db.conn.patchByFields TEST_TABLE, fields, Map.of('foo', 'green') // no exception = pass
    }

    static void byContainsMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def contains = Map.of 'value', 'another'
        db.conn.patchByContains TEST_TABLE, contains, Map.of('numValue', 12)
        def doc = db.conn.findFirstByContains TEST_TABLE, contains, JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'two', doc.get().id, 'The incorrect document was returned'
        assertEquals 12, doc.get().numValue, 'The document was not updated'
    }

    static void byContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def contains = Map.of 'value', 'updated'
        assertFalse db.conn.existsByContains(TEST_TABLE, contains), 'There should be no matching documents'
        db.conn.patchByContains TEST_TABLE, contains, Map.of('sub.foo', 'green') // no exception = pass
    }

    static void byJsonPathMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def path = '$.numValue ? (@ > 10)'
        db.conn.patchByJsonPath TEST_TABLE, path, Map.of('value', 'blue')
        def docs = db.conn.findByJsonPath TEST_TABLE, path, JsonDocument
        assertEquals 2, docs.size(), 'There should have been two documents returned'
        docs.forEach {
            assertTrue List.of('four', 'five').contains(it.id), "An incorrect document was returned (${it.id})"
            assertEquals 'blue', it.value, "The value for ID ${it.id} was incorrect"
        }
    }

    static void byJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def path = '$.numValue ? (@ > 100)'
        assertFalse(db.conn.existsByJsonPath(TEST_TABLE, path),
                'There should be no documents with numeric values over 100')
        db.conn.patchByJsonPath TEST_TABLE, path, Map.of('value', 'blue') // no exception = pass
    }
}
