package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.Configuration
import java.io.PrintWriter
import solutions.bitbadger.documents.java.Custom as CoreCustom
import java.sql.Connection
import java.sql.ResultSet

/**
 * Custom query execution functions
 */
object Custom {

    /**
     * Execute a query that returns a list of results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     */
    inline fun <reified TDoc : Any> list(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        mapFunc: (ResultSet) -> TDoc
    ) = Parameters.apply(conn, query, parameters).use { Results.toCustomList(it, mapFunc) }

    /**
     * Execute a query that returns a list of results (creates connection)
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     */
    inline fun <reified TDoc : Any> list(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        mapFunc: (ResultSet) -> TDoc
    ) = Configuration.dbConn().use { list(query, parameters, it, mapFunc) }

    /**
     * Execute a query that returns a JSON array of results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function to extract the JSON from the query
     * @return A JSON array of results for the given query
     * @throws DocumentException If parameters are invalid
     */
    fun jsonArray(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        mapFunc: (ResultSet) -> String
    ) = CoreCustom.jsonArray(query, parameters, conn, mapFunc)

    /**
     * Execute a query that returns a JSON array of results (creates connection)
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function to extract the JSON from the query
     * @return A JSON array of results for the given query
     * @throws DocumentException If parameters are invalid
     */
    fun jsonArray(query: String, parameters: Collection<Parameter<*>> = listOf(), mapFunc: (ResultSet) -> String) =
        CoreCustom.jsonArray(query, parameters, mapFunc)

    /**
     * Execute a query, writing its JSON array of results to the given `PrintWriter`
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param writer The writer to which the results should be written
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function to extract the JSON from the query
     * @throws DocumentException If parameters are invalid
     */
    fun writeJsonArray(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        writer: PrintWriter,
        conn: Connection,
        mapFunc: (ResultSet) -> String
    ) = CoreCustom.writeJsonArray(query, parameters, writer, conn, mapFunc)

    /**
     * Execute a query, writing its JSON array of results to the given `PrintWriter` (creates connection)
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param writer The writer to which the results should be written
     * @param mapFunc The mapping function to extract the JSON from the query
     * @throws DocumentException If parameters are invalid
     */
    fun writeJsonArray(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        writer: PrintWriter,
        mapFunc: (ResultSet) -> String
    ) = CoreCustom.writeJsonArray(query, parameters, writer, mapFunc)

    /**
     * Execute a query that returns one or no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return The document if one matches the query, `null` otherwise
     */
    inline fun <reified TDoc : Any> single(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        mapFunc: (ResultSet) -> TDoc
    ) = list("$query LIMIT 1", parameters, conn, mapFunc).singleOrNull()

    /**
     * Execute a query that returns one or no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return The document if one matches the query, `null` otherwise
     */
    inline fun <reified TDoc : Any> single(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        noinline mapFunc: (ResultSet) -> TDoc
    ) = Configuration.dbConn().use { single(query, parameters, it, mapFunc) }

    /**
     * Execute a query that returns JSON for one or no documents
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return The JSON for the document if found, an empty object (`{}`) if not
     * @throws DocumentException If parameters are invalid
     */
    fun jsonSingle(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        mapFunc: (ResultSet) -> String
    ) = CoreCustom.jsonSingle(query, parameters, conn, mapFunc)

    /**
     * Execute a query that returns JSON for one or no documents (creates connection)
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return The JSON for the document if found, an empty object (`{}`) if not
     * @throws DocumentException If parameters are invalid
     */
    fun jsonSingle(query: String, parameters: Collection<Parameter<*>> = listOf(), mapFunc: (ResultSet) -> String) =
        CoreCustom.jsonSingle(query, parameters, mapFunc)

    /**
     * Execute a query that returns no results
     *
     * @param query The query to retrieve the results
     * @param conn The connection over which the query should be executed
     * @param parameters Parameters to use for the query
     */
    fun nonQuery(query: String, parameters: Collection<Parameter<*>> = listOf(), conn: Connection) =
        CoreCustom.nonQuery(query, parameters, conn)

    /**
     * Execute a query that returns no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     */
    fun nonQuery(query: String, parameters: Collection<Parameter<*>> = listOf()) =
        Configuration.dbConn().use { nonQuery(query, parameters, it) }

    /**
     * Execute a query that returns a scalar result
     *
     * @param query The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     */
    inline fun <reified T : Any> scalar(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        mapFunc: (ResultSet) -> T
    ) = Parameters.apply(conn, query, parameters).use { stmt ->
        stmt.executeQuery().use { rs ->
            rs.next()
            mapFunc(rs)
        }
    }

    /**
     * Execute a query that returns a scalar result
     *
     * @param query The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     */
    inline fun <reified T : Any> scalar(
        query: String, parameters: Collection<Parameter<*>> = listOf(), mapFunc: (ResultSet) -> T
    ) = Configuration.dbConn().use { scalar(query, parameters, it, mapFunc) }
}
