package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.java.Delete as JvmDelete
import solutions.bitbadger.documents.kotlinx.extensions.*
import solutions.bitbadger.documents.query.DeleteQuery
import java.sql.Connection

/**
 * Functions to delete documents
 */
object Delete {

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId The ID of the document to be deleted
     * @param conn The connection on which the deletion should be executed
     */
    fun <TKey> byId(tableName: String, docId: TKey, conn: Connection) =
        JvmDelete.byId(tableName, docId, conn)

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId The ID of the document to be deleted
     */
    fun <TKey> byId(tableName: String, docId: TKey) =
        JvmDelete.byId(tableName, docId)

    /**
     * Delete documents using a field comparison
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param conn The connection on which the deletion should be executed
     */
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null, conn: Connection) =
        JvmDelete.byFields(tableName, fields, howMatched, conn)

    /**
     * Delete documents using a field comparison
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     */
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        JvmDelete.byFields(tableName, fields, howMatched)

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(tableName: String, criteria: TContains, conn: Connection) =
        conn.customNonQuery(DeleteQuery.byContains(tableName), listOf(Parameters.json<TContains>(":criteria", criteria)))

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria The object for which JSON containment should be checked
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(tableName: String, criteria: TContains) =
        Configuration.dbConn().use { byContains(tableName, criteria, it) }

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path The JSON path comparison to match
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String, conn: Connection) =
        JvmDelete.byJsonPath(tableName, path, conn)

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path The JSON path comparison to match
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String) =
        JvmDelete.byJsonPath(tableName, path)
}
