package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.AutoId
import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.kotlinx.extensions.customNonQuery
import solutions.bitbadger.documents.query.DocumentQuery
import solutions.bitbadger.documents.query.Where
import solutions.bitbadger.documents.query.statementWhere
import java.sql.Connection

/**
 * Functions for manipulating documents
 */
object Document {

    /**
     * Insert a new document
     *
     * @param tableName The table into which the document should be inserted (may include schema)
     * @param document The document to be inserted
     * @param conn The connection on which the query should be executed
     */
    inline fun <reified TDoc> insert(tableName: String, document: TDoc, conn: Connection) {
        val strategy = Configuration.autoIdStrategy
        val query = if (strategy == AutoId.DISABLED) {
            DocumentQuery.insert(tableName)
        } else {
            val idField   = Configuration.idField
            val dialect   = Configuration.dialect("Create auto-ID insert query")
            val dataParam = if (AutoId.needsAutoId(strategy, document, idField)) {
                when (dialect) {
                    Dialect.POSTGRESQL ->
                        when (strategy) {
                            AutoId.NUMBER        -> "' || (SELECT coalesce(max(data->>'$idField')::numeric, 0) + 1 " +
                                                            "FROM $tableName) || '"
                            AutoId.UUID          -> "\"${AutoId.generateUUID()}\""
                            AutoId.RANDOM_STRING -> "\"${AutoId.generateRandomString()}\""
                            else                 -> "\"' || (:data)->>'$idField' || '\""
                        }.let { ":data::jsonb || ('{\"$idField\":$it}')::jsonb" }

                    Dialect.SQLITE ->
                        when (strategy) {
                            AutoId.NUMBER        -> "(SELECT coalesce(max(data->>'$idField'), 0) + 1 FROM $tableName)"
                            AutoId.UUID          -> "'${AutoId.generateUUID()}'"
                            AutoId.RANDOM_STRING -> "'${AutoId.generateRandomString()}'"
                            else                 -> "(:data)->>'$idField'"
                        }.let { "json_set(:data, '$.$idField', $it)" }
                }
            } else {
                ":data"
            }

            DocumentQuery.insert(tableName).replace(":data", dataParam)
        }
        conn.customNonQuery(query, listOf(Parameters.json(":data", document)))
    }

    /**
     * Insert a new document
     *
     * @param tableName The table into which the document should be inserted (may include schema)
     * @param document The document to be inserted
     */
    inline fun <reified TDoc> insert(tableName: String, document: TDoc) =
        Configuration.dbConn().use { insert(tableName, document, it) }

    /**
     * Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table in which the document should be saved (may include schema)
     * @param document The document to be saved
     * @param conn The connection on which the query should be executed
     */
    inline fun <reified TDoc> save(tableName: String, document: TDoc, conn: Connection) =
        conn.customNonQuery(DocumentQuery.save(tableName), listOf(Parameters.json(":data", document)))

    /**
     * Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table in which the document should be saved (may include schema)
     * @param document The document to be saved
     */
    inline fun <reified TDoc> save(tableName: String, document: TDoc) =
        Configuration.dbConn().use { save(tableName, document, it) }

    /**
     * Update (replace) a document by its ID
     *
     * @param tableName The table in which the document should be replaced (may include schema)
     * @param docId The ID of the document to be replaced
     * @param document The document to be replaced
     * @param conn The connection on which the query should be executed
     */
    inline fun <TKey, reified TDoc> update(tableName: String, docId: TKey, document: TDoc, conn: Connection) =
        conn.customNonQuery(
            statementWhere(DocumentQuery.update(tableName), Where.byId(":id", docId)),
            Parameters.addFields(
                listOf(Field.equal(Configuration.idField, docId, ":id")),
                mutableListOf(Parameters.json(":data", document))
            )
        )

    /**
     * Update (replace) a document by its ID
     *
     * @param tableName The table in which the document should be replaced (may include schema)
     * @param docId The ID of the document to be replaced
     * @param document The document to be replaced
     */
    inline fun <TKey, reified TDoc> update(tableName: String, docId: TKey, document: TDoc) =
        Configuration.dbConn().use { update(tableName, docId, document, it) }
}
