package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.Parameter
import solutions.bitbadger.documents.ParameterType
import solutions.bitbadger.documents.java.Parameters as JvmParameters
import java.sql.Connection

/**
 * Functions to assist with the creation and implementation of parameters for SQL queries
 *
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 */
object Parameters {

    /**
     * Assign parameter names to any fields that do not have them assigned
     *
     * @param fields The collection of fields to be named
     * @return The collection of fields with parameter names assigned
     */
    fun nameFields(fields: Collection<Field<*>>): Collection<Field<*>> =
        JvmParameters.nameFields(fields)

    /**
     * Create a parameter by encoding a JSON object
     *
     * @param name The parameter name
     * @param value The object to be encoded as JSON
     * @return A parameter with the value encoded
     */
    inline fun <reified T> json(name: String, value: T) =
        Parameter(name, ParameterType.JSON, DocumentConfig.serialize<T>(value))

    /**
     * Add field parameters to the given set of parameters
     *
     * @param fields The fields being compared in the query
     * @param existing Any existing parameters for the query (optional, defaults to empty collection)
     * @return A collection of parameters for the query
     */
    fun addFields(fields: Collection<Field<*>>, existing: MutableCollection<Parameter<*>> = mutableListOf()) =
        JvmParameters.addFields(fields, existing)

    /**
     * Replace the parameter names in the query with question marks
     *
     * @param query The query with named placeholders
     * @param parameters The parameters for the query
     * @return The query, with name parameters changed to `?`s
     */
    fun replaceNamesInQuery(query: String, parameters: Collection<Parameter<*>>) =
        JvmParameters.replaceNamesInQuery(query, parameters)

    /**
     * Apply the given parameters to the given query, returning a prepared statement
     *
     * @param conn The active JDBC connection
     * @param query The query
     * @param parameters The parameters for the query
     * @return A `PreparedStatement` with the parameter names replaced with `?` and parameter values bound
     * @throws DocumentException If parameter names are invalid or number value types are invalid
     */
    fun apply(conn: Connection, query: String, parameters: Collection<Parameter<*>>) =
        JvmParameters.apply(conn, query, parameters)

    /**
     * Create parameters for field names to be removed from a document
     *
     * @param names The names of the fields to be removed
     * @param parameterName The parameter name to use for the query
     * @return A list of parameters to use for building the query
     * @throws DocumentException If the dialect has not been set
     */
    fun fieldNames(names: Collection<String>, parameterName: String = ":name") =
        JvmParameters.fieldNames(names, parameterName)
}
