package solutions.bitbadger.documents.kotlinx.tests.integration

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import solutions.bitbadger.documents.kotlinx.extensions.*
import solutions.bitbadger.documents.kotlinx.tests.*
import java.io.PrintWriter
import java.io.StringWriter
import kotlin.test.assertEquals
import kotlin.test.assertTrue

/**
 * Tests for the JSON-returning functions
 *
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
object JsonFunctions {

    /**
     * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
     * This function will do a crude string replacement to match the target string based on the dialect being tested.
     *
     * @param json The JSON which should be returned
     * @return The actual expected JSON based on the database being tested
     */
    fun maybeJsonB(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> json
            Dialect.POSTGRESQL -> json.replace("\":", "\": ").replace(",\"", ", \"")
        }

    /**
     * Create a snippet of JSON to find a document ID
     *
     * @param id The ID of the document
     * @return A connection-aware ID to check for presence and positioning
     */
    private fun docId(id: String) =
        maybeJsonB("{\"id\":\"$id\"")

    private fun checkAllDefault(json: String) {
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> {
                assertTrue(json.contains(JsonDocument.one), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.two), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.three), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found in JSON ($json)")
            }
            Dialect.POSTGRESQL -> {
                assertTrue(json.contains(docId("one")), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(docId("two")), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(docId("three")), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(docId("four")), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(docId("five")), "Document 'five' not found in JSON ($json)")
            }
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun allDefault(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkAllDefault(db.conn.jsonAll(TEST_TABLE))
    }

    fun writeAllDefault(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonAll(TEST_TABLE, writer)
        checkAllDefault(output.toString())
    }

    private fun checkAllEmpty(json: String) =
        assertEquals("[]", json, "There should have been no documents returned")

    fun allEmpty(db: ThrowawayDatabase) =
        checkAllEmpty(db.conn.jsonAll(TEST_TABLE))

    fun writeAllEmpty(db: ThrowawayDatabase) {
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonAll(TEST_TABLE, writer)
        checkAllEmpty(output.toString())
    }

    private fun checkByIdString(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.two, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("two")), "An incorrect document was returned ($json)")
        }

    fun byIdString(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByIdString(db.conn.jsonById(TEST_TABLE, "two"))
    }

    fun writeByIdString(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonById(TEST_TABLE, writer, "two")
        checkByIdString(output.toString())
    }

    private fun checkByIdNumber(json: String) =
        assertEquals(
            maybeJsonB("""{"key":18,"text":"howdy"}"""),
            json,
            "The document should have been found by numeric ID"
        )

    fun byIdNumber(db: ThrowawayDatabase) {
        Configuration.idField = "key"
        try {
            db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
            checkByIdNumber(db.conn.jsonById(TEST_TABLE, 18))
        } finally {
            Configuration.idField = "id"
        }
    }

    fun writeByIdNumber(db: ThrowawayDatabase) {
        Configuration.idField = "key"
        try {
            db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
            val output = StringWriter()
            val writer = PrintWriter(output)
            db.conn.writeJsonById(TEST_TABLE, writer, 18)
            checkByIdNumber(output.toString())
        } finally {
            Configuration.idField = "id"
        }
    }

    private fun checkByIdNotFound(json: String) =
        assertEquals("{}", json, "There should have been no document returned")

    fun byIdNotFound(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByIdNotFound(db.conn.jsonById(TEST_TABLE, "x"))
    }

    fun writeByIdNotFound(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonById(TEST_TABLE, writer, "x")
        checkByIdNotFound(output.toString())
    }

    private fun checkByFieldsMatch(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals("[${JsonDocument.four}]", json, "The incorrect document was returned")
            Dialect.POSTGRESQL -> {
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId("four")),"The incorrect document was returned ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }

    fun byFieldsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByFieldsMatch(
            db.conn.jsonByFields(
                TEST_TABLE, listOf(Field.any("value", listOf("blue", "purple")), Field.exists("sub")), FieldMatch.ALL
            )
        )
    }

    fun writeByFieldsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByFields(
            TEST_TABLE,
            writer,
            listOf(Field.any("value", listOf("blue", "purple")), Field.exists("sub")),
            FieldMatch.ALL
        )
        checkByFieldsMatch(output.toString())
    }

    private fun checkByFieldsMatchOrdered(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(
                "[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly"
            )

            Dialect.POSTGRESQL -> {
                val fiveIdx = json.indexOf(docId("five"))
                val fourIdx = json.indexOf(docId("four"))
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }

    fun byFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByFieldsMatchOrdered(
            db.conn.jsonByFields(
                TEST_TABLE, listOf(Field.equal("value", "purple")), orderBy = listOf(Field.named("id"))
            )
        )
    }

    fun writeByFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByFields(
            TEST_TABLE, writer, listOf(Field.equal("value", "purple")), orderBy = listOf(Field.named("id"))
        )
        checkByFieldsMatchOrdered(output.toString())
    }

    private fun checkByFieldsMatchNumIn(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals("[${JsonDocument.three}]", json, "The incorrect document was returned")
            Dialect.POSTGRESQL -> {
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId("three")), "The incorrect document was returned ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }

    fun byFieldsMatchNumIn(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByFieldsMatchNumIn(db.conn.jsonByFields(TEST_TABLE, listOf(Field.any("numValue", listOf(2, 4, 6, 8)))))
    }

    fun writeByFieldsMatchNumIn(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, listOf(Field.any("numValue", listOf(2, 4, 6, 8))))
        checkByFieldsMatchNumIn(output.toString())
    }

    private fun checkByFieldsNoMatch(json: String) =
        assertEquals("[]", json, "There should have been no documents returned")

    fun byFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByFieldsNoMatch(db.conn.jsonByFields(TEST_TABLE, listOf(Field.greater("numValue", 100))))
    }

    fun writeByFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, listOf(Field.greater("numValue", 100)))
        checkByFieldsNoMatch(output.toString())
    }

    private fun checkByFieldsMatchInArray(json: String) {
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        assertTrue(json.contains(docId("first")), "The 'first' document was not found ($json)")
        assertTrue(json.contains(docId("second")), "The 'second' document was not found ($json)")
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun byFieldsMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        checkByFieldsMatchInArray(
            db.conn.jsonByFields(TEST_TABLE, listOf(Field.inArray("values", TEST_TABLE, listOf("c"))))
        )
    }

    fun writeByFieldsMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, listOf(Field.inArray("values", TEST_TABLE, listOf("c"))))
        checkByFieldsMatchInArray(output.toString())
    }

    private fun checkByFieldsNoMatchInArray(json: String) =
        assertEquals("[]", json, "There should have been no documents returned")

    fun byFieldsNoMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        checkByFieldsNoMatchInArray(
            db.conn.jsonByFields(TEST_TABLE, listOf(Field.inArray("values", TEST_TABLE, listOf("j"))))
        )
    }

    fun writeByFieldsNoMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByFields(TEST_TABLE, writer, listOf(Field.inArray("values", TEST_TABLE, listOf("j"))))
        checkByFieldsNoMatchInArray(output.toString())
    }

    private fun checkByContainsMatch(json: String) {
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> {
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
            }
            Dialect.POSTGRESQL -> {
                assertTrue(json.contains(docId("four")), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId("five")), "Document 'five' not found ($json)")
            }
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun byContainsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByContainsMatch(db.conn.jsonByContains<Map<String, String>>(TEST_TABLE, mapOf("value" to "purple")))
    }

    fun writeByContainsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByContains<Map<String, String>>(TEST_TABLE, writer, mapOf("value" to "purple"))
        checkByContainsMatch(output.toString())
    }

    private fun checkByContainsMatchOrdered(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(
                "[${JsonDocument.two},${JsonDocument.four}]", json, "The documents were not ordered correctly"
            )
            Dialect.POSTGRESQL -> {
                val twoIdx = json.indexOf(docId("two"))
                val fourIdx = json.indexOf(docId("four"))
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(twoIdx >= 0, "Document 'two' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(twoIdx < fourIdx, "Document 'two' should have been before 'four' ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }

    fun byContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByContainsMatchOrdered(
            db.conn.jsonByContains<Map<String, Map<String, String>>>(
                TEST_TABLE, mapOf("sub" to mapOf("foo" to "green")), listOf(Field.named("value"))
            )
        )
    }

    fun writeByContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByContains<Map<String, Map<String, String>>>(
            TEST_TABLE, writer, mapOf("sub" to mapOf("foo" to "green")), listOf(Field.named("value"))
        )
        checkByContainsMatchOrdered(output.toString())
    }

    private fun checkByContainsNoMatch(json: String) =
        assertEquals("[]", json, "There should have been no documents returned")

    fun byContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByContainsNoMatch(db.conn.jsonByContains<Map<String, String>>(TEST_TABLE, mapOf("value" to "indigo")))
    }

    fun writeByContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByContains<Map<String, String>>(TEST_TABLE, writer, mapOf("value" to "indigo"))
        checkByContainsNoMatch(output.toString())
    }

    private fun checkByJsonPathMatch(json: String) {
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> {
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
            }
            Dialect.POSTGRESQL -> {
                assertTrue(json.contains(docId("four")), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId("five")), "Document 'five' not found ($json)")
            }
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun byJsonPathMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByJsonPathMatch(db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)"))
    }

    fun writeByJsonPathMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)")
        checkByJsonPathMatch(output.toString())
    }

    private fun checkByJsonPathMatchOrdered(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(
                "[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly"
            )

            Dialect.POSTGRESQL -> {
                val fiveIdx = json.indexOf(docId("five"))
                val fourIdx = json.indexOf(docId("four"))
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }

    fun byJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByJsonPathMatchOrdered(
            db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", listOf(Field.named("id")))
        )
    }

    fun writeByJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)", listOf(Field.named("id")))
        checkByJsonPathMatchOrdered(output.toString())
    }

    private fun checkByJsonPathNoMatch(json: String) =
        assertEquals("[]", json, "There should have been no documents returned")

    fun byJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkByJsonPathNoMatch(db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"))
    }

    fun writeByJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 100)")
        checkByJsonPathNoMatch(output.toString())
    }

    private fun checkFirstByFieldsMatchOne(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.two, json, "The incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("two")), "The incorrect document was returned ($json)")
        }

    fun firstByFieldsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByFieldsMatchOne(db.conn.jsonFirstByFields(TEST_TABLE, listOf(Field.equal("value", "another"))))
    }

    fun writeFirstByFieldsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, listOf(Field.equal("value", "another")))
        checkFirstByFieldsMatchOne(output.toString())
    }

    private fun checkFirstByFieldsMatchMany(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertTrue(
                json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
                "Expected document 'two' or 'four' ($json)"
            )
            Dialect.POSTGRESQL -> assertTrue(
                json.contains(docId("two")) || json.contains(docId("four")),
                "Expected document 'two' or 'four' ($json)"
            )
        }

    fun firstByFieldsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByFieldsMatchMany(db.conn.jsonFirstByFields(TEST_TABLE, listOf(Field.equal("sub.foo", "green"))))
    }

    fun writeFirstByFieldsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, listOf(Field.equal("sub.foo", "green")))
        checkFirstByFieldsMatchMany(output.toString())
    }

    private fun checkFirstByFieldsMatchOrdered(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.four, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("four")), "An incorrect document was returned ($json)")
        }

    fun firstByFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByFieldsMatchOrdered(
            db.conn.jsonFirstByFields(
                TEST_TABLE, listOf(Field.equal("sub.foo", "green")), orderBy = listOf(Field.named("n:numValue DESC"))
            )
        )
    }

    fun writeFirstByFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByFields(
            TEST_TABLE,
            writer,
            listOf(Field.equal("sub.foo", "green")),
            orderBy = listOf(Field.named("n:numValue DESC"))
        )
        checkFirstByFieldsMatchOrdered(output.toString())
    }

    private fun checkFirstByFieldsNoMatch(json: String) =
        assertEquals("{}", json, "There should have been no document returned")

    fun firstByFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByFieldsNoMatch(db.conn.jsonFirstByFields(TEST_TABLE, listOf(Field.equal("value", "absent"))))
    }

    fun writeFirstByFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByFields(TEST_TABLE, writer, listOf(Field.equal("value", "absent")))
        checkFirstByFieldsNoMatch(output.toString())
    }

    private fun checkFirstByContainsMatchOne(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.one, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("one")), "An incorrect document was returned ($json)")
        }

    fun firstByContainsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByContainsMatchOne(
            db.conn.jsonFirstByContains<Map<String, String>>(TEST_TABLE, mapOf("value" to "FIRST!"))
        )
    }

    fun writeFirstByContainsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByContains<Map<String, String>>(TEST_TABLE, writer, mapOf("value" to "FIRST!"))
    }

    private fun checkFirstByContainsMatchMany(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertTrue(
                json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                "Expected document 'four' or 'five' ($json)"
            )
            Dialect.POSTGRESQL -> assertTrue(
                json.contains(docId("four")) || json.contains(docId("five")),
                "Expected document 'four' or 'five' ($json)"
            )
        }

    fun firstByContainsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByContainsMatchMany(
            db.conn.jsonFirstByContains<Map<String, String>>(TEST_TABLE, mapOf("value" to "purple"))
        )
    }

    fun writeFirstByContainsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByContains<Map<String, String>>(TEST_TABLE, writer, mapOf("value" to "purple"))
        checkFirstByContainsMatchMany(output.toString())
    }

    private fun checkFirstByContainsMatchOrdered(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.five, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("five")), "An incorrect document was returned ($json)")
        }

    fun firstByContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByContainsMatchOrdered(
            db.conn.jsonFirstByContains<Map<String, String>>(
                TEST_TABLE, mapOf("value" to "purple"), listOf(Field.named("sub.bar NULLS FIRST"))
            )
        )
    }

    fun writeFirstByContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByContains<Map<String, String>>(
            TEST_TABLE, writer, mapOf("value" to "purple"), listOf(Field.named("sub.bar NULLS FIRST"))
        )
        checkFirstByContainsMatchOrdered(output.toString())
    }

    private fun checkFirstByContainsNoMatch(json: String) =
        assertEquals("{}", json, "There should have been no document returned")

    fun firstByContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByContainsNoMatch(
            db.conn.jsonFirstByContains<Map<String, String>>(TEST_TABLE, mapOf("value" to "indigo"))
        )
    }

    fun writeFirstByContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByContains<Map<String, String>>(TEST_TABLE, writer, mapOf("value" to "indigo"))
        checkFirstByContainsNoMatch(output.toString())
    }

    private fun checkFirstByJsonPathMatchOne(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.two, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("two")), "An incorrect document was returned ($json)")
        }

    fun firstByJsonPathMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByJsonPathMatchOne(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ == 10)"))
    }

    fun writeFirstByJsonPathMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ == 10)")
        checkFirstByJsonPathMatchOne(output.toString())
    }

    private fun checkFirstByJsonPathMatchMany(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertTrue(
                json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                "Expected document 'four' or 'five' ($json)"
            )
            Dialect.POSTGRESQL -> assertTrue(
                json.contains(docId("four")) || json.contains(docId("five")),
                "Expected document 'four' or 'five' ($json)"
            )
        }

    fun firstByJsonPathMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByJsonPathMatchMany(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)"))
    }

    fun writeFirstByJsonPathMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)")
        checkFirstByJsonPathMatchMany(output.toString())
    }

    private fun checkFirstByJsonPathMatchOrdered(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.four, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("four")), "An incorrect document was returned ($json)")
        }

    fun firstByJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByJsonPathMatchOrdered(
            db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", listOf(Field.named("id DESC")))
        )
    }

    fun writeFirstByJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)", listOf(Field.named("id DESC")))
        checkFirstByJsonPathMatchOrdered(output.toString())
    }

    private fun checkFirstByJsonPathNoMatch(json: String) =
        assertEquals("{}", json, "There should have been no document returned")

    fun firstByJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        checkFirstByJsonPathNoMatch(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"))
    }

    fun writeFirstByJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 100)")
        checkFirstByJsonPathNoMatch(output.toString())
    }
}
