package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Configuration, Parameter}

import java.io.PrintWriter
import java.sql.{Connection, ResultSet}
import scala.reflect.ClassTag
import scala.util.Using

object Custom:

  /**
   * Execute a query that returns a list of results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, parameters: Seq[Parameter[?]], conn: Connection,
                mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): List[Doc] =
    Using(Parameters.apply(conn, query, parameters)) { stmt => Results.toCustomList[Doc](stmt, mapFunc) }.get

  /**
   * Execute a query that returns a list of results
   *
   * @param query   The query to retrieve the results
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, conn: Connection, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
               (using tag: ClassTag[Doc]): List[Doc] =
    list(query, Nil, conn, mapFunc)

  /**
   * Execute a query that returns a list of results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
               (using tag: ClassTag[Doc]): List[Doc] =
    Using(Configuration.dbConn()) { conn => list[Doc](query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns a list of results (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): List[Doc] =
    list(query, List(), mapFunc)

  /**
   * Execute a query that returns a JSON array of results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def jsonArray(query: String, parameters: Seq[Parameter[?]], conn: Connection, mapFunc: ResultSet => String): String =
    Using(Parameters.apply(conn, query, parameters)) { stmt => Results.toJsonArray(stmt, mapFunc) }.get

  /**
   * Execute a query that returns a JSON array of results
   *
   * @param query   The query to retrieve the results
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def jsonArray(query: String, conn: Connection, mapFunc: ResultSet => String): String =
    jsonArray(query, Nil, conn, mapFunc)

  /**
   * Execute a query that returns a JSON array of results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def jsonArray(query: String, parameters: Seq[Parameter[?]], mapFunc: ResultSet => String): String =
    Using(Configuration.dbConn()) { conn => jsonArray(query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns a JSON array of results (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def jsonArray(query: String, mapFunc: ResultSet => String): String =
    jsonArray(query, Nil, mapFunc)

  /**
   * Execute a query that writes a JSON array of results to the given `PrintWriter`
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param writer     The writer to which the results should be written
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def writeJsonArray(query: String, parameters: Seq[Parameter[?]], writer: PrintWriter, conn: Connection,
                     mapFunc: ResultSet => String): Unit =
    Using(Parameters.apply(conn, query, parameters)) { stmt => Results.writeJsonArray(writer, stmt, mapFunc) }

  /**
   * Execute a query that returns a JSON array of results
   *
   * @param query   The query to retrieve the results
   * @param writer  The writer to which the results should be written
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def writeJsonArray(query: String, conn: Connection, writer: PrintWriter, mapFunc: ResultSet => String): Unit =
    writeJsonArray(query, Nil, writer, conn, mapFunc)

  /**
   * Execute a query that returns a JSON array of results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param writer     The writer to which the results should be written
   * @param mapFunc    The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def writeJsonArray(query: String, parameters: Seq[Parameter[?]], writer: PrintWriter,
                     mapFunc: ResultSet => String): Unit =
    Using(Configuration.dbConn()) { conn => writeJsonArray(query, parameters, writer, conn, mapFunc) }

  /**
   * Execute a query that returns a JSON array of results (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param writer  The writer to which the results should be written
   * @param mapFunc The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def writeJsonArray(query: String, writer: PrintWriter, mapFunc: ResultSet => String): Unit =
    writeJsonArray(query, Nil, writer, mapFunc)

  /**
   * Execute a query that returns one or no results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, parameters: Seq[Parameter[?]], conn: Connection,
                  mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): Option[Doc] =
    list[Doc](s"$query LIMIT 1", parameters, conn, mapFunc).headOption

  /**
   * Execute a query that returns one or no results
   *
   * @param query   The query to retrieve the results
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, conn: Connection, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                 (using tag: ClassTag[Doc]): Option[Doc] =
    list[Doc](s"$query LIMIT 1", List(), conn, mapFunc).headOption

  /**
   * Execute a query that returns one or no results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                 (using tag: ClassTag[Doc]): Option[Doc] =
    Using(Configuration.dbConn()) { conn => single[Doc](query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns one or no results (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): Option[Doc] =
    single[Doc](query, List(), mapFunc)

  /**
   * Execute a query that returns JSON for one or no documents
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The JSON for the document if found, an empty object (`{}`) if not
   * @throws DocumentException If parameters are invalid
   */
  def jsonSingle(query: String, parameters: Seq[Parameter[?]], conn: Connection, mapFunc: ResultSet => String): String =
    val result = jsonArray("$query LIMIT 1", parameters, conn, mapFunc)
    result match
      case "[]" => "{}"
      case _ => result.substring(1, result.length - 1)

  /**
   * Execute a query that returns JSON for one or no documents
   *
   * @param query   The query to retrieve the results
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return The JSON for the document if found, an empty object (`{}`) if not
   * @throws DocumentException If parameters are invalid
   */
  def jsonSingle(query: String, conn: Connection, mapFunc: ResultSet => String): String =
    jsonSingle(query, Nil, conn, mapFunc)

  /**
   * Execute a query that returns JSON for one or no documents (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The JSON for the document if found, an empty object (`{}`) if not
   * @throws DocumentException If parameters are invalid
   */
  def jsonSingle(query: String, parameters: Seq[Parameter[?]], mapFunc: ResultSet => String): String =
    Using(Configuration.dbConn()) { conn => jsonSingle(query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns JSON for one or no documents (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return The JSON for the document if found, an empty object (`{}`) if not
   * @throws DocumentException If parameters are invalid
   */
  def jsonSingle(query: String, mapFunc: ResultSet => String): String =
    jsonSingle(query, Nil, mapFunc)

  /**
   * Execute a query that returns no results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String, parameters: Seq[Parameter[?]], conn: Connection): Unit =
    Using(Parameters.apply(conn, query, parameters)) { stmt => stmt.executeUpdate() }

  /**
   * Execute a query that returns no results
   *
   * @param query The query to retrieve the results
   * @param conn  The connection over which the query should be executed
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String, conn: Connection): Unit =
    nonQuery(query, List(), conn)

  /**
   * Execute a query that returns no results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String, parameters: Seq[Parameter[?]]): Unit =
    Using(Configuration.dbConn()) { conn => nonQuery(query, parameters, conn) }

  /**
   * Execute a query that returns no results (creates connection)
   *
   * @param query The query to retrieve the results
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String): Unit =
    nonQuery(query, List())

  /**
   * Execute a query that returns a scalar result
   *
   * @param query      The query to retrieve the result
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, parameters: Seq[Parameter[?]], conn: Connection, mapFunc: (ResultSet, ClassTag[A]) => A)
               (using tag: ClassTag[A]): A =
    Using(Parameters.apply(conn, query, parameters)) { stmt =>
      Using(stmt.executeQuery()) { rs =>
        rs.next()
        mapFunc(rs, tag)
      }.get
    }.get

  /**
   * Execute a query that returns a scalar result
   *
   * @param query   The query to retrieve the result
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, conn: Connection, mapFunc: (ResultSet, ClassTag[A]) => A)(using tag: ClassTag[A]): A =
    scalar[A](query, List(), conn, mapFunc)

  /**
   * Execute a query that returns a scalar result (creates connection)
   *
   * @param query      The query to retrieve the result
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[A]) => A)
               (using tag: ClassTag[A]): A =
    Using(Configuration.dbConn()) { conn => scalar[A](query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns a scalar result (creates connection)
   *
   * @param query   The query to retrieve the result
   * @param mapFunc The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, mapFunc: (ResultSet, ClassTag[A]) => A)(using tag: ClassTag[A]): A =
    scalar[A](query, List(), mapFunc)
