package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Field, FieldMatch}
import solutions.bitbadger.documents.java.Delete as CoreDelete

import java.sql.Connection
import _root_.scala.jdk.CollectionConverters.*

/**
 * Functions to delete documents
 */
object Delete:

  /**
   * Delete a document by its ID
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param docId     The ID of the document to be deleted
   * @param conn      The connection on which the deletion should be executed
   * @throws DocumentException If no dialect has been configured
   */
  def byId[Key](tableName: String, docId: Key, conn: Connection): Unit =
    CoreDelete.byId(tableName, docId, conn)

  /**
  * Delete a document by its ID (creates connection)
  *
  * @param tableName The name of the table from which documents should be deleted
  * @param docId     The ID of the document to be deleted
  * @throws DocumentException If no connection string has been set
  */
  def byId[Key](tableName: String, docId: Key): Unit =
    CoreDelete.byId(tableName, docId)

  /**
   * Delete documents using a field comparison
   *
   * @param tableName  The name of the table from which documents should be deleted
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param conn       The connection on which the deletion should be executed
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch], conn: Connection): Unit =
    CoreDelete.byFields(tableName, fields.asJava, howMatched.orNull, conn)

  /**
   * Delete documents using a field comparison (creates connection)
   *
   * @param tableName  The name of the table from which documents should be deleted
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Unit =
    CoreDelete.byFields(tableName, fields.asJava, howMatched.orNull)

  /**
   * Delete documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param criteria  The object for which JSON containment should be checked
   * @param conn      The connection on which the deletion should be executed
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A, conn: Connection): Unit =
    CoreDelete.byContains(tableName, criteria, conn)

  /**
   * Delete documents using a JSON containment query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param criteria  The object for which JSON containment should be checked
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A): Unit =
    CoreDelete.byContains(tableName, criteria)

  /**
   * Delete documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param path      The JSON path comparison to match
   * @param conn      The connection on which the deletion should be executed
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String, conn: Connection): Unit =
    CoreDelete.byJsonPath(tableName, path, conn)

  /**
  * Delete documents using a JSON Path match query (PostgreSQL only; creates connection)
  *
  * @param tableName The name of the table from which documents should be deleted
  * @param path      The JSON path comparison to match
  * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
  */
  def byJsonPath(tableName: String, path: String): Unit =
    CoreDelete.byJsonPath(tableName, path)
