package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Field, Op, Parameter, ParameterName}
import solutions.bitbadger.documents.java.Parameters as CoreParameters

import java.sql.{Connection, PreparedStatement}
import java.util
import scala.collection.mutable
import scala.collection.mutable.ListBuffer
import _root_.scala.jdk.CollectionConverters.*

/**
 * Functions to assist with the creation and implementation of parameters for SQL queries
 */
object Parameters:

  /**
   * Assign parameter names to any fields that do not have them assigned
   *
   * @param fields The collection of fields to be named
   * @return The collection of fields with parameter names assigned
   */
  def nameFields(fields: Seq[Field[?]]): Seq[Field[?]] =
    val name = ParameterName()
    fields.map { it =>
      if (it.getParameterName == null || it.getParameterName.isEmpty)
          && !(Op.EXISTS :: Op.NOT_EXISTS :: Nil).contains(it.getComparison.getOp) then
        it.withParameterName(name.derive(null))
      else
        it
    }

  /**
   * Create a parameter by encoding a JSON object
   *
   * @param name  The parameter name
   * @param value The object to be encoded as JSON
   * @return A parameter with the value encoded
   */
  def json[A](name: String, value: A): Parameter[String] =
    CoreParameters.json(name, value)

  /**
   * Add field parameters to the given set of parameters
   *
   * @param fields   The fields being compared in the query
   * @param existing Any existing parameters for the query (optional, defaults to empty collection)
   * @return A collection of parameters for the query
   */
  def addFields(fields: Seq[Field[?]],
                existing: mutable.Buffer[Parameter[?]] = ListBuffer()): mutable.Buffer[Parameter[?]] =
    fields.foreach { it => it.appendParameter(new util.ArrayList[Parameter[?]]()).forEach(existing.append) }
    existing

  /**
   * Replace the parameter names in the query with question marks
   *
   * @param query      The query with named placeholders
   * @param parameters The parameters for the query
   * @return The query, with name parameters changed to `?`s
   */
  def replaceNamesInQuery(query: String, parameters: Seq[Parameter[?]]): String =
    CoreParameters.replaceNamesInQuery(query, parameters.asJava)

  /**
   * Apply the given parameters to the given query, returning a prepared statement
   *
   * @param conn       The active JDBC connection
   * @param query      The query
   * @param parameters The parameters for the query
   * @return A `PreparedStatement` with the parameter names replaced with `?` and parameter values bound
   * @throws DocumentException If parameter names are invalid or number value types are invalid
   */
  def apply(conn: Connection, query: String, parameters: Seq[Parameter[?]]): PreparedStatement =
    CoreParameters.apply(conn, query, parameters.asJava)

  /**
   * Create parameters for field names to be removed from a document
   *
   * @param names         The names of the fields to be removed
   * @param parameterName The parameter name to use for the query
   * @return A list of parameters to use for building the query
   * @throws DocumentException If the dialect has not been set
   */
  def fieldNames(names: Seq[String], parameterName: String = ":name"): mutable.Buffer[Parameter[?]] =
    CoreParameters.fieldNames(names.asJava, parameterName).asScala.toBuffer
