package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Field, FieldMatch}
import solutions.bitbadger.documents.java.Patch as CorePatch

import java.sql.Connection
import _root_.scala.jdk.CollectionConverters.*

/**
 * Functions to patch (partially update) documents
 */
object Patch:

  /**
   * Patch a document by its ID
   *
   * @param tableName The name of the table in which a document should be patched
   * @param docId     The ID of the document to be patched
   * @param patch     The object whose properties should be replaced in the document
   * @param conn      The connection on which the update should be executed
   * @throws DocumentException If no dialect has been configured
   */
  def byId[Key, Patch](tableName: String, docId: Key, patch: Patch, conn: Connection): Unit =
    CorePatch.byId(tableName, docId, patch, conn)

  /**
   * Patch a document by its ID (creates connection)
   *
   * @param tableName The name of the table in which a document should be patched
   * @param docId     The ID of the document to be patched
   * @param patch     The object whose properties should be replaced in the document
   * @throws DocumentException If no connection string has been set
   */
  def byId[Key, Patch](tableName: String, docId: Key, patch: Patch): Unit =
    CorePatch.byId(tableName, docId, patch)

  /**
   * Patch documents using a field comparison
   *
   * @param tableName  The name of the table in which documents should be patched
   * @param fields     The fields which should be compared
   * @param patch      The object whose properties should be replaced in the document
   * @param howMatched How the fields should be matched
   * @param conn       The connection on which the update should be executed
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields[Patch](tableName: String, fields: Seq[Field[?]], patch: Patch, howMatched: Option[FieldMatch],
                  conn: Connection): Unit =
    CorePatch.byFields(tableName, fields.asJava, patch, howMatched.orNull, conn)

  /**
   * Patch documents using a field comparison
   *
   * @param tableName The name of the table in which documents should be patched
   * @param fields    The fields which should be compared
   * @param patch     The object whose properties should be replaced in the document
   * @param conn      The connection on which the update should be executed
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields[Patch](tableName: String, fields: Seq[Field[?]], patch: Patch, conn: Connection): Unit =
    byFields(tableName, fields, patch, None, conn)

  /**
   * Patch documents using a field comparison (creates connection)
   *
   * @param tableName  The name of the table in which documents should be patched
   * @param fields     The fields which should be compared
   * @param patch      The object whose properties should be replaced in the document
   * @param howMatched How the fields should be matched
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def byFields[Patch](tableName: String, fields: Seq[Field[?]], patch: Patch,
                      howMatched: Option[FieldMatch] = None): Unit =
    CorePatch.byFields(tableName, fields.asJava, patch, howMatched.orNull)

  /**
   * Patch documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be patched
   * @param criteria  The object against which JSON containment should be checked
   * @param patch     The object whose properties should be replaced in the document
   * @param conn      The connection on which the update should be executed
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[A, Patch](tableName: String, criteria: A, patch: Patch, conn: Connection): Unit =
    CorePatch.byContains(tableName, criteria, patch, conn)

  /**
   * Patch documents using a JSON containment query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which documents should be patched
   * @param criteria  The object against which JSON containment should be checked
   * @param patch     The object whose properties should be replaced in the document
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byContains[A, Patch](tableName: String, criteria: A, patch: Patch): Unit =
    CorePatch.byContains(tableName, criteria, patch)

  /**
   * Patch documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be patched
   * @param path      The JSON path comparison to match
   * @param patch     The object whose properties should be replaced in the document
   * @param conn      The connection on which the update should be executed
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath[Patch](tableName: String, path: String, patch: Patch, conn: Connection): Unit =
    CorePatch.byJsonPath(tableName, path, patch, conn)

  /**
   * Patch documents using a JSON Path match query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which documents should be patched
   * @param path      The JSON path comparison to match
   * @param patch     The object whose properties should be replaced in the document
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byJsonPath[Patch](tableName: String, path: String, patch: Patch): Unit =
    CorePatch.byJsonPath(tableName, path, patch)
