package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Field, FieldMatch}
import solutions.bitbadger.documents.java.RemoveFields as CoreRemoveFields

import java.sql.Connection
import scala.jdk.CollectionConverters.*

/**
 * Functions to remove fields from documents
 */
object RemoveFields:

  /**
   * Remove fields from a document by its ID
   *
   * @param tableName The name of the table in which the document's fields should be removed
   * @param docId     The ID of the document to have fields removed
   * @param toRemove  The names of the fields to be removed
   * @param conn      The connection on which the update should be executed
   * @throws DocumentException If no dialect has been configured
   */
  def byId[Key](tableName: String, docId: Key, toRemove: Seq[String], conn: Connection): Unit =
    CoreRemoveFields.byId(tableName, docId, toRemove.asJava, conn)

  /**
   * Remove fields from a document by its ID (creates connection)
   *
   * @param tableName The name of the table in which the document's fields should be removed
   * @param docId     The ID of the document to have fields removed
   * @param toRemove  The names of the fields to be removed
   * @throws DocumentException If no connection string has been set
   */
  def byId[Key](tableName: String, docId: Key, toRemove: Seq[String]): Unit =
    CoreRemoveFields.byId(tableName, docId, toRemove.asJava)

  /**
   * Remove fields from documents using a field comparison
   *
   * @param tableName  The name of the table in which document fields should be removed
   * @param fields     The fields which should be compared
   * @param toRemove   The names of the fields to be removed
   * @param howMatched How the fields should be matched
   * @param conn       The connection on which the update should be executed
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], toRemove: Seq[String], howMatched: Option[FieldMatch],
               conn: Connection): Unit =
    CoreRemoveFields.byFields(tableName, fields.asJava, toRemove.asJava, howMatched.orNull, conn)

  /**
   * Remove fields from documents using a field comparison
   *
   * @param tableName  The name of the table in which document fields should be removed
   * @param fields     The fields which should be compared
   * @param toRemove   The names of the fields to be removed
   * @param conn       The connection on which the update should be executed
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], toRemove: Seq[String], conn: Connection): Unit =
    byFields(tableName, fields, toRemove, None, conn)

  /**
   * Remove fields from documents using a field comparison (creates connection)
   *
   * @param tableName  The name of the table in which document fields should be removed
   * @param fields     The fields which should be compared
   * @param toRemove   The names of the fields to be removed
   * @param howMatched How the fields should be matched
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], toRemove: Seq[String],
               howMatched: Option[FieldMatch] = None): Unit =
    CoreRemoveFields.byFields(tableName, fields.asJava, toRemove.asJava, howMatched.orNull)

  /**
   * Remove fields from documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document fields should be removed
   * @param criteria  The object against which JSON containment should be checked
   * @param toRemove  The names of the fields to be removed
   * @param conn      The connection on which the update should be executed
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A, toRemove: Seq[String], conn: Connection): Unit =
    CoreRemoveFields.byContains(tableName, criteria, toRemove.asJava, conn)

  /**
   * Remove fields from documents using a JSON containment query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which document fields should be removed
   * @param criteria  The object against which JSON containment should be checked
   * @param toRemove  The names of the fields to be removed
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A, toRemove: Seq[String]): Unit =
    CoreRemoveFields.byContains(tableName, criteria, toRemove.asJava)

  /**
   * Remove fields from documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document fields should be removed
   * @param path The JSON path comparison to match
   * @param toRemove The names of the fields to be removed
   * @param conn The connection on which the update should be executed
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String, toRemove: Seq[String], conn: Connection): Unit =
    CoreRemoveFields.byJsonPath(tableName, path, toRemove.asJava, conn)

  /**
   * Remove fields from documents using a JSON Path match query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which document fields should be removed
   * @param path The JSON path comparison to match
   * @param toRemove The names of the fields to be removed
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String, toRemove: Seq[String]): Unit =
    CoreRemoveFields.byJsonPath(tableName, path, toRemove.asJava)
