package solutions.bitbadger.documents.scala.extensions

import solutions.bitbadger.documents.{DocumentIndex, Field, FieldMatch, Parameter}
import solutions.bitbadger.documents.scala.*

import java.io.PrintWriter
import java.sql.{Connection, ResultSet}
import scala.reflect.ClassTag

extension (conn: Connection)

  // ~~~ CUSTOM QUERIES ~~~

  /**
   * Execute a query that returns a list of results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def customList[Doc](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                     (using tag: ClassTag[Doc]): List[Doc] =
    Custom.list[Doc](query, parameters, conn, mapFunc)

  /**
   * Execute a query that returns a list of results
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def customList[Doc](query: String, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                     (using tag: ClassTag[Doc]): List[Doc] =
    Custom.list[Doc](query, conn, mapFunc)

  /**
   * Execute a query that returns a JSON array of results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def customJsonArray(query: String, parameters: Seq[Parameter[?]], mapFunc: ResultSet => String): String =
    Custom.jsonArray(query, parameters, conn, mapFunc)

  /**
   * Execute a query that returns a JSON array of results
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def customJsonArray(query: String, mapFunc: ResultSet => String): String =
    Custom.jsonArray(query, mapFunc)

  /**
   * Execute a query that writes a JSON array of results to the given `PrintWriter`
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param writer     The writer to which the results should be written
   * @param mapFunc    The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def writeCustomJsonArray(query: String, parameters: Seq[Parameter[?]], writer: PrintWriter,
                           mapFunc: ResultSet => String): Unit =
    Custom.writeJsonArray(query, parameters, writer, conn, mapFunc)

  /**
   * Execute a query that writes a JSON array of results to the given `PrintWriter`
   *
   * @param query   The query to retrieve the results
   * @param writer  The writer to which the results should be written
   * @param mapFunc The mapping function to extract the JSON from the query
   * @return A JSON array of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def writeCustomJsonArray(query: String, writer: PrintWriter, mapFunc: ResultSet => String): Unit =
    Custom.writeJsonArray(query, writer, mapFunc)

  /**
   * Execute a query that returns one or no results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return An optional document, filled if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def customSingle[Doc](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                       (using tag: ClassTag[Doc]): Option[Doc] =
    Custom.single[Doc](query, parameters, conn, mapFunc)

  /**
   * Execute a query that returns one or no results
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return An optional document, filled if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def customSingle[Doc](query: String, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                       (using tag: ClassTag[Doc]): Option[Doc] =
    Custom.single[Doc](query, conn, mapFunc)

  /**
   * Execute a query that returns JSON for one or no documents (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The JSON for the document if found, an empty object (`{}`) if not
   * @throws DocumentException If parameters are invalid
   */
  def customJsonSingle(query: String, parameters: Seq[Parameter[?]], mapFunc: ResultSet => String): String =
    Custom.jsonSingle(query, parameters, conn, mapFunc)

  /**
   * Execute a query that returns JSON for one or no documents (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return The JSON for the document if found, an empty object (`{}`) if not
   * @throws DocumentException If parameters are invalid
   */
  def customJsonSingle(query: String, mapFunc: ResultSet => String): String =
    Custom.jsonSingle(query, mapFunc)

  /**
   * Execute a query that returns no results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @throws DocumentException If parameters are invalid
   */
  def customNonQuery(query: String, parameters: Seq[Parameter[?]] = Nil): Unit =
    Custom.nonQuery(query, parameters, conn)

  /**
   * Execute a query that returns a scalar result
   *
   * @param query      The query to retrieve the result
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def customScalar[A](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[A]) => A)
                     (using tag: ClassTag[A]): A =
    Custom.scalar[A](query, parameters, conn, mapFunc)

  /**
   * Execute a query that returns a scalar result
   *
   * @param query   The query to retrieve the result
   * @param mapFunc The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def customScalar[A](query: String, mapFunc: (ResultSet, ClassTag[A]) => A)(using tag: ClassTag[A]): A =
    Custom.scalar[A](query, conn, mapFunc)

  // ~~~ DEFINITION QUERIES ~~~

  /**
   * Create a document table if necessary
   *
   * @param tableName The table whose existence should be ensured (may include schema)
   * @throws DocumentException If the dialect is not configured
   */
  def ensureTable(tableName: String): Unit =
    Definition.ensureTable(tableName, conn)

  /**
   * Create an index on field(s) within documents in the specified table if necessary
   *
   * @param tableName The table to be indexed (may include schema)
   * @param indexName The name of the index to create
   * @param fields    One or more fields to be indexed<
   * @throws DocumentException If any dependent process does
   */
  def ensureFieldIndex(tableName: String, indexName: String, fields: Seq[String]): Unit =
    Definition.ensureFieldIndex(tableName, indexName, fields, conn)

  /**
   * Create a document index on a table (PostgreSQL only)
   *
   * @param tableName The table to be indexed (may include schema)
   * @param indexType The type of index to ensure
   * @throws DocumentException If called on a SQLite connection
   */
  def ensureDocumentIndex(tableName: String, indexType: DocumentIndex): Unit =
    Definition.ensureDocumentIndex (tableName, indexType, conn)

  // ~~~ DOCUMENT MANIPULATION QUERIES ~~~

  /**
   * Insert a new document
   *
   * @param tableName The table into which the document should be inserted (may include schema)
   * @param document  The document to be inserted
   * @throws DocumentException If IDs are misconfigured, or if the database command fails
   */
  def insert[Doc](tableName: String, document: Doc): Unit =
    Document.insert(tableName, document, conn)

  /**
   * Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
   *
   * @param tableName The table in which the document should be saved (may include schema)
   * @param document  The document to be saved
   * @throws DocumentException If the database command fails
   */
  def save[Doc](tableName: String, document: Doc): Unit =
    Document.save(tableName, document, conn)

  /**
   * Update (replace) a document by its ID
   *
   * @param tableName The table in which the document should be replaced (may include schema)
   * @param docId     The ID of the document to be replaced
   * @param document  The document to be replaced
   * @throws DocumentException If no dialect has been configured, or if the database command fails
   */
  def update[Key, Doc](tableName: String, docId: Key, document: Doc): Unit =
    Document.update(tableName, docId, document, conn)

  // ~~~ DOCUMENT COUNT QUERIES ~~~

  /**
   * Count all documents in the table
   *
   * @param tableName The name of the table in which documents should be counted
   * @return A count of the documents in the table
   * @throws DocumentException If any dependent process does
   */
  def countAll(tableName: String): Long =
    Count.all(tableName, conn)

  /**
   * Count documents using a field comparison
   *
   * @param tableName  The name of the table in which documents should be counted
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, default `ALL`)
   * @return A count of the matching documents in the table
   * @throws DocumentException If the dialect has not been configured
   */
  def countByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Long =
    Count.byFields(tableName, fields, howMatched, conn)

  /**
   * Count documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be counted
   * @param criteria  The object for which JSON containment should be checked
   * @return A count of the matching documents in the table
   * @throws DocumentException If called on a SQLite connection
   */
  def countByContains[A](tableName: String, criteria: A): Long =
    Count.byContains(tableName, criteria, conn)

  /**
   * Count documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be counted
   * @param path      The JSON path comparison to match
   * @return A count of the matching documents in the table
   * @throws DocumentException If called on a SQLite connection
   */
  def countByJsonPath(tableName: String, path: String): Long =
    Count.byJsonPath(tableName, path, conn)

  // ~~~ DOCUMENT EXISTENCE QUERIES ~~~

  /**
   * Determine a document's existence by its ID
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param docId     The ID of the document to be checked
   * @return True if the document exists, false if not
   * @throws DocumentException If no dialect has been configured
   */
  def existsById[Key](tableName: String, docId: Key): Boolean =
    Exists.byId(tableName, docId, conn)

  /**
   * Determine document existence using a field comparison
   *
   * @param tableName  The name of the table in which document existence should be checked
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def existsByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Boolean =
    Exists.byFields(tableName, fields, howMatched, conn)

  /**
   * Determine document existence using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param criteria  The object for which JSON containment should be checked
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If called on a SQLite connection
   */
  def existsByContains[A](tableName: String, criteria: A): Boolean =
    Exists.byContains(tableName, criteria, conn)

  /**
   * Determine document existence using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param path      The JSON path comparison to match
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If called on a SQLite connection
   */
  def existsByJsonPath(tableName: String, path: String): Boolean =
    Exists.byJsonPath(tableName, path, conn)

  // ~~~ DOCUMENT RETRIEVAL QUERIES (Domain Objects) ~~~

  /**
   * Retrieve all documents in the given table, ordering results by the optional given fields
   *
   * @param tableName The table from which documents should be retrieved
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents from the given table
   * @throws DocumentException If query execution fails
   */
  def findAll[Doc](tableName: String, orderBy: Seq[Field[?]] = Nil)(using tag: ClassTag[Doc]): List[Doc] =
    Find.all[Doc](tableName, orderBy, conn)

  /**
   * Retrieve a document by its ID
   *
   * @param tableName The table from which the document should be retrieved
   * @param docId     The ID of the document to retrieve
   * @return The document if it is found, `None` otherwise
   * @throws DocumentException If no dialect has been configured
   */
  def findById[Key, Doc](tableName: String, docId: Key)(using tag: ClassTag[Doc]): Option[Doc] =
    Find.byId[Key, Doc](tableName, docId, conn)

  /**
   * Retrieve documents using a field comparison, ordering results by the optional given fields
   *
   * @param tableName  The table from which the document should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents matching the field comparison
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def findByFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                        orderBy: Seq[Field[?]] = Nil)(using tag: ClassTag[Doc]): List[Doc] =
    Find.byFields[Doc](tableName, fields, howMatched, orderBy, conn)

  /**
   * Retrieve documents using a JSON containment query, ordering results by the optional given fields (PostgreSQL
   * only)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents matching the JSON containment query
   * @throws DocumentException If called on a SQLite connection
   */
  def findByContains[Doc, A](tableName: String, criteria: A, orderBy: Seq[Field[?]] = Nil)
                            (using tag: ClassTag[Doc]): List[Doc] =
    Find.byContains[Doc, A](tableName, criteria, orderBy, conn)

  /**
   * Retrieve documents using a JSON Path match query, ordering results by the optional given fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents matching the JSON Path match query
   * @throws DocumentException If called on a SQLite connection
   */
  def findByJsonPath[Doc](tableName: String, path: String, orderBy: Seq[Field[?]] = Nil)
                         (using tag: ClassTag[Doc]): List[Doc] =
    Find.byJsonPath[Doc](tableName, path, orderBy, conn)

  /**
   * Retrieve the first document using a field comparison and optional ordering fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return The first document matching the field comparison, or `None` if no matches are found
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def findFirstByFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                             orderBy: Seq[Field[?]] = Nil)(using tag: ClassTag[Doc]): Option[Doc] =
    Find.firstByFields[Doc](tableName, fields, howMatched, orderBy, conn)

  /**
   * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return The first document matching the JSON containment query, or `None` if no matches are found
   * @throws DocumentException If called on a SQLite connection
   */
  def findFirstByContains[Doc, A](tableName: String, criteria: A, orderBy: Seq[Field[?]] = Nil)
                                 (using tag: ClassTag[Doc]): Option[Doc] =
    Find.firstByContains[Doc, A](tableName, criteria, orderBy, conn)

  /**
   * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return The first document matching the JSON Path match query, or `None` if no matches are found
   * @throws DocumentException If called on a SQLite connection
   */
  def findFirstByJsonPath[Doc](tableName: String, path: String, orderBy: Seq[Field[?]] = Nil)
                              (using tag: ClassTag[Doc]): Option[Doc] =
    Find.firstByJsonPath[Doc](tableName, path, orderBy, conn)

  // ~~~ DOCUMENT RETRIEVAL QUERIES (Raw JSON) ~~~

  /**
   * Retrieve all documents in the given table
   *
   * @param tableName The table from which documents should be retrieved
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A JSON array of documents from the given table
   * @throws DocumentException If no connection string has been set, or if query execution fails
   */
  def jsonAll(tableName: String, orderBy: Seq[Field[?]] = Nil): String =
    Json.all(tableName, orderBy, conn)

  /**
   * Retrieve a document by its ID
   *
   * @param tableName The table from which the document should be retrieved
   * @param docId     The ID of the document to retrieve
   * @return A JSON document if found, an empty JSON object if not found
   * @throws DocumentException If no connection string has been set
   */
  def jsonById[Key](tableName: String, docId: Key): String =
    Json.byId(tableName, docId, conn)

  /**
   * Retrieve documents using a field comparison, ordering results by the given fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A JSON array of documents matching the field comparison
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def jsonByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                   orderBy: Seq[Field[?]] = Nil): String =
    Json.byFields(tableName, fields, howMatched, orderBy, conn)

  /**
   * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A JSON array of documents matching the JSON containment query
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def jsonByContains[A](tableName: String, criteria: A, orderBy: Seq[Field[?]] = Nil): String =
    Json.byContains(tableName, criteria, orderBy, conn)

  /**
   * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A JSON array of documents matching the JSON Path match query
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def jsonByJsonPath(tableName: String, path: String, orderBy: Seq[Field[?]] = Nil): String =
    Json.byJsonPath(tableName, path, orderBy, conn)

  /**
   * Retrieve the first document using a field comparison and optional ordering fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return The first JSON document matching the field comparison if found, an empty JSON object otherwise
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def jsonFirstByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                        orderBy: Seq[Field[?]] = Nil): String =
    Json.firstByFields(tableName, fields, howMatched, orderBy, conn)

  /**
   * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return The first JSON document matching the JSON containment query if found, an empty JSON object otherwise
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def jsonFirstByContains[A](tableName: String, criteria: A, orderBy: Seq[Field[?]] = Nil): String =
    Json.firstByContains(tableName, criteria, orderBy, conn)

  /**
   * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return The first JSON document matching the JSON Path match query if found, an empty JSON object otherwise
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def jsonFirstByJsonPath(tableName: String, path: String, orderBy: Seq[Field[?]] = Nil): String =
    Json.firstByJsonPath(tableName, path, orderBy, conn)

  // ~~~ DOCUMENT RETRIEVAL QUERIES (Write raw JSON to PrintWriter) ~~~

  /**
   * Write all documents in the given table to the given `PrintWriter`, ordering results by the optional given fields
   *
   * @param tableName The table from which documents should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if query execution fails
   */
  def writeJsonAll(tableName: String, writer: PrintWriter, orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeAll(tableName, writer, orderBy, conn)

  /**
   * Write a document to the given `PrintWriter` by its ID
   *
   * @param tableName The table from which the document should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param docId     The ID of the document to retrieve
   * @throws DocumentException If no connection string has been set
   */
  def writeJsonById[Key](tableName: String, writer: PrintWriter, docId: Key): Unit =
    Json.writeById(tableName, writer, docId, conn)

  /**
   * Write documents to the given `PrintWriter` using a field comparison, ordering results by the given fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param writer     The `PrintWriter` to which the results should be written
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def writeJsonByFields(tableName: String, writer: PrintWriter, fields: Seq[Field[?]],
                        howMatched: Option[FieldMatch] = None, orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeByFields(tableName, writer, fields, howMatched, orderBy, conn)

  /**
   * Write documents to the given `PrintWriter` using a JSON containment query, ordering results by the given fields
   * (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def writeJsonByContains[A](tableName: String, writer: PrintWriter, criteria: A, orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeByContains(tableName, writer, criteria, orderBy, conn)

  /**
   * Write documents to the given `PrintWriter` using a JSON Path match query, ordering results by the given fields
   * (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def writeJsonByJsonPath(tableName: String, writer: PrintWriter, path: String, orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeByJsonPath(tableName, writer, path, orderBy, conn)

  /**
   * Write the first document to the given `PrintWriter` using a field comparison and ordering fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def writeJsonFirstByFields(tableName: String, writer: PrintWriter, fields: Seq[Field[?]],
                             howMatched: Option[FieldMatch] = None, orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeFirstByFields(tableName, writer, fields, howMatched, orderBy, conn)

  /**
   * Write the first document to the given `PrintWriter` using a JSON containment query and ordering fields (PostgreSQL
   * only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def writeJsonFirstByContains[A](tableName: String, writer: PrintWriter, criteria: A,
                                  orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeFirstByContains(tableName, writer, criteria, orderBy, conn)

  /**
   * Write the first document to the given `PrintWriter` using a JSON Path match query and ordering fields (PostgreSQL
   * only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param writer    The `PrintWriter` to which the results should be written
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def writeJsonFirstByJsonPath(tableName: String, writer: PrintWriter, path: String,
                               orderBy: Seq[Field[?]] = Nil): Unit =
    Json.writeFirstByJsonPath(tableName, writer, path, orderBy, conn)

  // ~~~ DOCUMENT PATCH (PARTIAL UPDATE) QUERIES ~~~

  /**
   * Patch a document by its ID
   *
   * @param tableName The name of the table in which a document should be patched
   * @param docId     The ID of the document to be patched
   * @param patch     The object whose properties should be replaced in the document
   * @throws DocumentException If no dialect has been configured
   */
  def patchById[Key, Patch](tableName: String, docId: Key, patch: Patch): Unit =
    Patch.byId(tableName, docId, patch, conn)

  /**
   * Patch documents using a field comparison
   *
   * @param tableName  The name of the table in which documents should be patched
   * @param fields     The fields which should be compared
   * @param patch      The object whose properties should be replaced in the document
   * @param howMatched How the fields should be matched
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def patchByFields[Patch](tableName: String, fields: Seq[Field[?]], patch: Patch,
                           howMatched: Option[FieldMatch] = None): Unit =
    Patch.byFields(tableName, fields, patch, howMatched, conn)

  /**
  * Patch documents using a JSON containment query (PostgreSQL only)
  *
  * @param tableName The name of the table in which documents should be patched
  * @param criteria  The object against which JSON containment should be checked
  * @param patch     The object whose properties should be replaced in the document
  * @throws DocumentException If called on a SQLite connection
  */
  def patchByContains[A, Patch](tableName: String, criteria: A, patch: Patch): Unit =
    Patch.byContains(tableName, criteria, patch, conn)

  /**
   * Patch documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be patched
   * @param path The JSON path comparison to match
   * @param patch The object whose properties should be replaced in the document
   * @throws DocumentException If called on a SQLite connection
   */
  def patchByJsonPath[Patch](tableName: String, path: String, patch: Patch): Unit =
    Patch.byJsonPath(tableName, path, patch, conn)

  // ~~~ DOCUMENT FIELD REMOVAL QUERIES ~~~

  /**
   * Remove fields from a document by its ID
   *
   * @param tableName The name of the table in which the document's fields should be removed
   * @param docId     The ID of the document to have fields removed
   * @param toRemove  The names of the fields to be removed
   * @throws DocumentException If no dialect has been configured
   */
  def removeFieldsById[Key](tableName: String, docId: Key, toRemove: Seq[String]): Unit =
    RemoveFields.byId(tableName, docId, toRemove, conn)

  /**
   * Remove fields from documents using a field comparison
   *
   * @param tableName  The name of the table in which document fields should be removed
   * @param fields     The fields which should be compared
   * @param toRemove   The names of the fields to be removed
   * @param howMatched How the fields should be matched
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def removeFieldsByFields(tableName: String, fields: Seq[Field[?]], toRemove: Seq[String],
                           howMatched: Option[FieldMatch] = None): Unit =
    RemoveFields.byFields(tableName, fields, toRemove, howMatched, conn)

  /**
   * Remove fields from documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document fields should be removed
   * @param criteria  The object against which JSON containment should be checked
   * @param toRemove  The names of the fields to be removed
   * @throws DocumentException If called on a SQLite connection
   */
  def removeFieldsByContains[A](tableName: String, criteria: A, toRemove: Seq[String]): Unit =
    RemoveFields.byContains(tableName, criteria, toRemove, conn)

  /**
  * Remove fields from documents using a JSON Path match query (PostgreSQL only)
  *
  * @param tableName The name of the table in which document fields should be removed
  * @param path      The JSON path comparison to match
  * @param toRemove  The names of the fields to be removed
  * @throws DocumentException If called on a SQLite connection
  */
  def removeFieldsByJsonPath(tableName: String, path: String, toRemove: Seq[String]): Unit =
    RemoveFields.byJsonPath(tableName, path, toRemove, conn)

  // ~~~ DOCUMENT DELETION QUERIES ~~~

  /**
   * Delete a document by its ID
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param docId     The ID of the document to be deleted
   * @throws DocumentException If no dialect has been configured
   */
  def deleteById[Key](tableName: String, docId: Key): Unit =
    Delete.byId(tableName, docId, conn)

  /**
   * Delete documents using a field comparison
   *
   * @param tableName  The name of the table from which documents should be deleted
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def deleteByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Unit =
    Delete.byFields(tableName, fields, howMatched, conn)

  /**
   * Delete documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param criteria  The object for which JSON containment should be checked
   * @throws DocumentException If called on a SQLite connection
   */
  def deleteByContains[A](tableName: String, criteria: A): Unit =
    Delete.byContains(tableName, criteria, conn)

  /**
   * Delete documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table from which documents should be deleted
   * @param path      The JSON path comparison to match
   * @throws DocumentException If called on a SQLite connection
   */
  def deleteByJsonPath(tableName: String, path: String): Unit =
    Delete.byJsonPath(tableName, path, conn)
