package solutions.bitbadger.documents.scala.tests

import org.junit.jupiter.api.Assertions.*
import org.junit.jupiter.api.{AfterEach, DisplayName, Test}
import solutions.bitbadger.documents.{Dialect, Field, FieldMatch}
import solutions.bitbadger.documents.query.QueryUtils

import scala.jdk.CollectionConverters.*

@DisplayName("Scala | Query | Package Functions")
class QueryUtilsTest:

  /**
   * Clear the connection string (resets Dialect)
   */
  @AfterEach
  def cleanUp(): Unit =
    ForceDialect.none()

  @Test
  @DisplayName("statementWhere generates correctly")
  def statementWhere(): Unit =
    assertEquals("x WHERE y", QueryUtils.statementWhere("x", "y"), "Statements not combined correctly")

  @Test
  @DisplayName("byId generates a numeric ID query | PostgreSQL")
  def byIdNumericPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("test WHERE (data->>'id')::numeric = :id", QueryUtils.byId("test", 9))

  @Test
  @DisplayName("byId generates an alphanumeric ID query | PostgreSQL")
  def byIdAlphaPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("unit WHERE data->>'id' = :id", QueryUtils.byId("unit", "18"))

  @Test
  @DisplayName("byId generates ID query | SQLite")
  def byIdSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("yo WHERE data->>'id' = :id", QueryUtils.byId("yo", 27))

  @Test
  @DisplayName("byFields generates default field query | PostgreSQL")
  def byFieldsMultipleDefaultPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value",
      QueryUtils.byFields("this", List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava))

  @Test
  @DisplayName("byFields generates default field query | SQLite")
  def byFieldsMultipleDefaultSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("this WHERE data->>'a' = :the_a AND data->>'b' = :b_value",
      QueryUtils.byFields("this", List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava))

  @Test
  @DisplayName("byFields generates ANY field query | PostgreSQL")
  def byFieldsMultipleAnyPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value",
      QueryUtils.byFields("that", List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava,
        FieldMatch.ANY))

  @Test
  @DisplayName("byFields generates ANY field query | SQLite")
  def byFieldsMultipleAnySQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("that WHERE data->>'a' = :the_a OR data->>'b' = :b_value",
      QueryUtils.byFields("that", List(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")).asJava,
        FieldMatch.ANY))

  @Test
  @DisplayName("orderBy generates for no fields")
  def orderByNone(): Unit =
    assertEquals("", QueryUtils.orderBy(List().asJava, Dialect.POSTGRESQL),
      "ORDER BY should have been blank (PostgreSQL)")
    assertEquals("", QueryUtils.orderBy(List().asJava, Dialect.SQLITE), "ORDER BY should have been blank (SQLite)")

  @Test
  @DisplayName("orderBy generates single, no direction | PostgreSQL")
  def orderBySinglePostgres(): Unit =
    assertEquals(" ORDER BY data->>'TestField'",
      QueryUtils.orderBy(List(Field.named("TestField")).asJava, Dialect.POSTGRESQL),
      "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates single, no direction | SQLite")
  def orderBySingleSQLite(): Unit =
    assertEquals(" ORDER BY data->>'TestField'",
      QueryUtils.orderBy(List(Field.named("TestField")).asJava, Dialect.SQLITE),
      "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates multiple with direction | PostgreSQL")
  def orderByMultiplePostgres(): Unit =
    assertEquals(" ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
      QueryUtils.orderBy(
        List(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")).asJava,
        Dialect.POSTGRESQL),
      "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates multiple with direction | SQLite")
  def orderByMultipleSQLite(): Unit =
    assertEquals(" ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
      QueryUtils.orderBy(
        List(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")).asJava,
        Dialect.SQLITE),
      "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates numeric ordering | PostgreSQL")
  def orderByNumericPostgres(): Unit =
    assertEquals(" ORDER BY (data->>'Test')::numeric",
      QueryUtils.orderBy(List(Field.named("n:Test")).asJava, Dialect.POSTGRESQL), "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates numeric ordering | SQLite")
  def orderByNumericSQLite(): Unit =
    assertEquals(" ORDER BY data->>'Test'", QueryUtils.orderBy(List(Field.named("n:Test")).asJava, Dialect.SQLITE),
      "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates case-insensitive ordering | PostgreSQL")
  def orderByCIPostgres(): Unit =
    assertEquals(" ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
      QueryUtils.orderBy(List(Field.named("i:Test.Field DESC NULLS FIRST")).asJava, Dialect.POSTGRESQL),
      "ORDER BY not constructed correctly")

  @Test
  @DisplayName("orderBy generates case-insensitive ordering | SQLite")
  def orderByCISQLite(): Unit =
    assertEquals(" ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
      QueryUtils.orderBy(List(Field.named("i:Test.Field ASC NULLS LAST")).asJava, Dialect.SQLITE),
      "ORDER BY not constructed correctly")
